import sys
import os
import re
import getdist
import numpy as np

sys.path.append("../autolens/")

import galaxy
from profiles import light_profiles, mass_profiles
from astropy import cosmology

def getModelFolders(dir):
    regex = re.compile('.*Hyper')
    folders = [name for name in os.listdir(dir) if os.path.isdir(dir+name) and re.match(regex, name) == None]
    return folders

ltm_skip = 0
center_skip = 0

image_dir = '/gpfs/weighted_data/pdtw24/PL_Data/SL03_2/'  # Dir of Object to make evidence tables from

image_name = 'SLACSJ0252+0039'
redshift = 0.2803
source_redshift = 0.9818
scale_r = 6.83924
einstein_radius = 1.02
source_light_min = 0.8
source_light_max = 1.3
slacs_mass = '{0:e}'.format(10**11.25)
slacs_radius = 4.4

image_name = 'SLACSJ1250+0523'
redshift = 0.2318
source_redshift = 0.7953
scale_r = 7.8660
kpc_to_arc = 3.8138
einstein_radius = 1.120
source_light_min = 0.5
source_light_max = 1.5
slacs_mass =  '{0:e}'.format(10**11.26)
slacs_radius = 4.18

# image_name = 'SLACSJ1430+4105'
# redshift = 0.2850
# source_redshift = 0.5753
# scale_r = 6.771
# kpc_to_arc = 4.4309
# einstein_radius = 1.468
# source_light_min = 0.2
# source_light_max = 2.3
# slacs_mass =  '{0:e}'.format(10**11.73)
# slacs_radius = 6.53

# pipeline_folder = 'Pipeline_Total'
# phase_folder = 'PL_Phase_8'

pipeline_folder = 'Pipeline_Total'
phase_folder = 'PL_Phase_8'

image_dir = image_dir + image_name + '/' + pipeline_folder + '/' + phase_folder + '/'

model_folder = getModelFolders(image_dir)

pdf_file = image_dir + model_folder[0] + '/' + image_name + '.txt'

pdf = getdist.mcsamples.loadMCSamples(pdf_file)

values = pdf.getMeans()
params = pdf.getParams()

upper_conf = []
lower_conf = []
for i in range(len(values)):
    upper_conf.append(pdf.confidence(paramVec=i, limfrac=0.999, upper=True))
    lower_conf.append(pdf.confidence(paramVec=i, limfrac=0.999, upper=False))

if image_name == 'SLACSJ1250+0523' or image_name == 'SLACSJ1430+4105':
    center_skip = 2

sersic_bulge = light_profiles.SersicLightProfile(centre=(values[0], values[1]), axis_ratio=values[5], phi=values[6], intensity=values[2], effective_radius=values[3],
                                               sersic_index=values[4])

setattr(sersic_bulge, 'centre_lower_limit_3_sigma', (lower_conf[0], lower_conf[1]))
setattr(sersic_bulge, 'axis_ratio_lower_limit_3_sigma', lower_conf[5])
setattr(sersic_bulge, 'phi_lower_limit_3_sigma', lower_conf[6])
setattr(sersic_bulge, 'intensity_lower_limit_3_sigma', lower_conf[2])
setattr(sersic_bulge, 'effective_radius_lower_limit_3_sigma', lower_conf[3])
setattr(sersic_bulge, 'sersic_index_lower_limit_3_sigma', lower_conf[4])

setattr(sersic_bulge, 'centre_upper_limit_3_sigma', (upper_conf[0], upper_conf[1]))
setattr(sersic_bulge, 'axis_ratio_upper_limit_3_sigma', upper_conf[5])
setattr(sersic_bulge, 'phi_upper_limit_3_sigma', upper_conf[6])
setattr(sersic_bulge, 'intensity_upper_limit_3_sigma', upper_conf[2])
setattr(sersic_bulge, 'effective_radius_upper_limit_3_sigma', upper_conf[3])
setattr(sersic_bulge, 'sersic_index_upper_limit_3_sigma', upper_conf[4])

exponential_halo = light_profiles.ExponentialLightProfile(axis_ratio=values[9+center_skip], phi=values[10+center_skip], intensity=values[7+center_skip],
                                                        effective_radius=values[8+center_skip])

setattr(exponential_halo, 'centre_lower_limit_3_sigma', (lower_conf[9+center_skip], lower_conf[10+center_skip]))
setattr(exponential_halo, 'axis_ratio_lower_limit_3_sigma', lower_conf[9+center_skip])
setattr(exponential_halo, 'phi_lower_limit_3_sigma', lower_conf[10+center_skip])
setattr(exponential_halo, 'intensity_lower_limit_3_sigma', lower_conf[7+center_skip])
setattr(exponential_halo, 'effective_radius_lower_limit_3_sigma', lower_conf[8+center_skip])
setattr(exponential_halo, 'centre_upper_limit_3_sigma', (upper_conf[9+center_skip], upper_conf[10+center_skip]))
setattr(exponential_halo, 'axis_ratio_upper_limit_3_sigma', upper_conf[9+center_skip])
setattr(exponential_halo, 'phi_upper_limit_3_sigma', upper_conf[10+center_skip])
setattr(exponential_halo, 'intensity_upper_limit_3_sigma', upper_conf[7+center_skip])
setattr(exponential_halo, 'effective_radius_upper_limit_3_sigma', upper_conf[8+center_skip])

power_law_mass = mass_profiles.EllipticalPowerLawMassProfile(centre=(values[11+center_skip], values[12+center_skip]),
                                                             einstein_radius=values[13+center_skip],
                                                             axis_ratio=values[14+center_skip], phi=values[15+center_skip],
                                                             slope=values[16+center_skip])

# power_law_mass = mass_profiles.EllipticalPowerLawMassProfile(centre=(values[10], values[11]),
#                                                              einstein_radius=values[12],
#                                                              axis_ratio=values[13], phi=values[14],
#                                                              slope=values[15])

setattr(power_law_mass, 'centre_lower_limit_3_sigma', (upper_conf[11+center_skip], upper_conf[12+center_skip]))
setattr(power_law_mass, 'einstein_radius_lower_limit_3_sigma', lower_conf[13+center_skip])
setattr(power_law_mass, 'axis_ratio_lower_limit_3_sigma', lower_conf[14+center_skip])
setattr(power_law_mass ,'phi_lower_limit_3_sigma', lower_conf[15+center_skip])
setattr(power_law_mass ,'slope_lower_limit_3_sigma', lower_conf[16+center_skip])

setattr(power_law_mass, 'centre_upper_limit_3_sigma', (upper_conf[11+center_skip], upper_conf[12+center_skip]))
setattr(power_law_mass, 'einstein_radius_upper_limit_3_sigma', upper_conf[13+center_skip])
setattr(power_law_mass, 'axis_ratio_upper_limit_3_sigma', upper_conf[14+center_skip])
setattr(power_law_mass ,'phi_upper_limit_3_sigma', upper_conf[15+center_skip])
setattr(power_law_mass ,'slope_upper_limit_3_sigma', upper_conf[16+center_skip])

lens_galaxy=galaxy.Galaxy(redshift=redshift, light_profiles=[sersic_bulge, exponential_halo],
                          mass_profiles=[power_law_mass])

source_galaxy = galaxy.Galaxy(redshift=source_redshift)

galaxy.LensingPlanes(galaxies=[lens_galaxy, source_galaxy], cosmological_model=cosmology.LambdaCDM(H0=70, Om0=0.3, Ode0=0.7))

radius = lens_galaxy.mass_profiles[0].einstein_radius

total_mass = lens_galaxy.mass_within_circles(radius=radius)

lens_galaxy.mass_profiles[0].centre = lens_galaxy.mass_profiles[0].centre_lower_limit_3_sigma
lens_galaxy.mass_profiles[0].einstein_radius = lens_galaxy.mass_profiles[0].einstein_radius_lower_limit_3_sigma
lens_galaxy.mass_profiles[0].phi = lens_galaxy.mass_profiles[0].phi_lower_limit_3_sigma
lens_galaxy.mass_profiles[0].slope = lens_galaxy.mass_profiles[0].slope_lower_limit_3_sigma

total_mass_lower = lens_galaxy.mass_within_circles(radius=radius)

lens_galaxy.mass_profiles[0].centre = lens_galaxy.mass_profiles[0].centre_upper_limit_3_sigma
lens_galaxy.mass_profiles[0].einstein_radius = lens_galaxy.mass_profiles[0].einstein_radius_upper_limit_3_sigma
lens_galaxy.mass_profiles[0].phi = lens_galaxy.mass_profiles[0].phi_upper_limit_3_sigma
lens_galaxy.mass_profiles[0].slope = lens_galaxy.mass_profiles[0].slope_upper_limit_3_sigma

total_mass_upper = lens_galaxy.mass_within_circles(radius=radius)

# print(lens_galaxy.mass_profiles[0].centre)
# print(lens_galaxy.mass_profiles[0].einstein_radius*kpc_to_arc)
# print(lens_galaxy.mass_profiles[0].axis_ratio)
# print(lens_galaxy.mass_profiles[0].phi)
# print(lens_galaxy.mass_profiles[0].slope)
# stop

# print(lens_galaxy.critical_density)

# print(slacs_mass)
# print(slacs_radius)


# einstein_radius = slacs_radius*lens_galaxy.arcsec_per_kpc

print(einstein_radius)
print('Our Mass = ', '{0:e}'.format(total_mass), '(', '{0:e}'.format(total_mass_upper), '{0:e}'.format(total_mass_lower), ')')
print(float(slacs_mass) / total_mass )

stop

lens_galaxy.plot_density_as_function_of_radius(maximum_radius=5.0, number_bins=300, image_name=image_name, plot_errors=True,
                                               labels=['Sersic Bulge', 'EllipticalExponentialMass Stellar Halo', 'Dark Matter Halo'])


from autolens import exc

from astropy import constants
import math
import numpy as np


class AbstractTracer(object):
    @property
    def all_planes(self):
        raise NotImplementedError()

    @property
    def galaxy_images(self):
        """
        Returns
        -------
        galaxy_images: [ndarray]
            An image for each galaxy in this ray tracer
        """
        return [galaxy_image for plane in self.all_planes for galaxy_image in plane.galaxy_images]

    @property
    def hyper_galaxies(self):
        return [hyper_galaxy for plane in self.all_planes for hyper_galaxy in
                plane.hyper_galaxies]

    @property
    def galaxies(self):
        return [galaxy for plane in self.all_planes for galaxy in plane.galaxies]

    @property
    def all_with_hyper_galaxies(self):
        return len(list(filter(None, self.hyper_galaxies))) == len(self.galaxies)


class Tracer(AbstractTracer):

    @property
    def all_planes(self):
        return [self.image_plane, self.source_plane]

    def __init__(self, lens_galaxies, source_galaxies, image_plane_grids, cosmology=None):
        """The ray-tracing calculations, defined by a lensing system with just one image-plane and source-plane.

        By default, this has no associated cosmology, thus all calculations are performed in arc seconds and galaxies \
        do not need input redshifts. For computational efficiency, it is recommend this ray-tracing class is used for \
        lens modeling, provided cosmological information is not necessary.

        If a cosmology is supplied, the plane's angular diameter distances, conversion factors, etc. will be computed.

        Parameters
        ----------
        lens_galaxies : [Galaxy]
            The list of lens galaxies in the image-plane.
        source_galaxies : [Galaxy]
            The list of source galaxies in the source-plane.
        image_plane_grids : mask.GridCollection
            The image-plane coordinate grids where ray-tracing calculation are performed, (this includes the
            image-grid, sub-grid, blurring-grid, etc.).
        cosmology : astropy.cosmology.Planck15
            The cosmology of the ray-tracing calculation.
        """
        if cosmology is not None:
            self.geometry = TracerGeometry(redshifts=[lens_galaxies[0].redshift, source_galaxies[0].redshift],
                                           cosmology=cosmology)
        else:
            self.geometry = None

        self.image_plane = Plane(lens_galaxies, image_plane_grids, compute_deflections=True)

        source_plane_grids = self.image_plane.trace_to_next_plane()

        self.source_plane = Plane(source_galaxies, source_plane_grids, compute_deflections=False)

    def generate_image_of_galaxy_light_profiles(self):
        """Generate the image of the galaxies over the entire ray trace."""
        return self.image_plane.generate_image_of_galaxy_light_profiles(
        ) + self.source_plane.generate_image_of_galaxy_light_profiles()

    def generate_blurring_image_of_galaxy_light_profiles(self):
        """Generate the image of all galaxy light profiles in the blurring regions of the image."""
        return self.image_plane.blurring_image_from_galaxy_light_profiles(
        ) + self.source_plane.blurring_image_from_galaxy_light_profiles()

    def reconstructors_from_source_plane(self, borders, cluster_mask):
        return self.source_plane.reconstructor_from_plane(borders, cluster_mask)


class MultiTracer(AbstractTracer):

    @property
    def all_planes(self):
        return self.planes

    def __init__(self, galaxies, image_plane_grids, cosmology):
        """The ray-tracing calculations, defined by a lensing system with just one image-plane and source-plane.

        This has no associated cosmology, thus all calculations are performed in arc seconds and galaxies do not need
        known redshift measurements. For computational efficiency, it is recommend this ray-tracing class is used for
        lens modeling, provided cosmological information is not necessary.

        Parameters
        ----------
        galaxies : [Galaxy]
            The list of galaxies in the ray-tracing calculation.
        image_plane_grids : mask.GridCollection
            The image-plane coordinate grids where ray-tracing calculation are performed, (this includes the
            image-grid, sub-grid, blurring-grid, etc.).
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """

        self.galaxies_redshift_order = sorted(galaxies, key=lambda galaxy: galaxy.redshift, reverse=False)

        # Ideally we'd extract the planes_red_Shfit order from the list above. However, I dont know how to extract it
        # Using a list of class attributes so make a list of redshifts for now.

        galaxy_redshifts = list(map(lambda galaxy: galaxy.redshift, self.galaxies_redshift_order))
        self.planes_redshift_order = [redshift for i, redshift in enumerate(galaxy_redshifts)
                                      if redshift not in galaxy_redshifts[:i]]
        self.geometry = TracerGeometry(redshifts=self.planes_redshift_order, cosmology=cosmology)

        # TODO : Idea is to get a list of all galaxies in each plane - can you clean up the logic below?

        self.planes_galaxies = []

        for (plane_index, plane_redshift) in enumerate(self.planes_redshift_order):
            self.planes_galaxies.append(list(map(lambda galaxy:
                                                 galaxy if galaxy.redshift == plane_redshift else None,
                                                 self.galaxies_redshift_order)))
            self.planes_galaxies[plane_index] = list(filter(None, self.planes_galaxies[plane_index]))

        self.planes = []

        for plane_index in range(0, len(self.planes_redshift_order)):

            if plane_index < len(self.planes_redshift_order) - 1:
                compute_deflections = True
            elif plane_index == len(self.planes_redshift_order) - 1:
                compute_deflections = False
            else:
                raise exc.RayTracingException('A galaxy was not correctly allocated its previous / next redshifts')

            new_grid = image_plane_grids

            if plane_index > 0:
                for previous_plane_index in range(plane_index):
                    scaling_factor = self.geometry.scaling_factor(plane_i=previous_plane_index,
                                                                  plane_j=plane_index)

                    def scale(grid):
                        return np.multiply(scaling_factor, grid)

                    scaled_deflections = self.planes[previous_plane_index].deflections. \
                        apply_function(scale)

                    def subtract_scaled_deflections(grid, scaled_deflection):
                        return np.subtract(grid, scaled_deflection)

                    new_grid = new_grid.map_function(subtract_scaled_deflections, scaled_deflections)

            self.planes.append(Plane(galaxies=self.planes_galaxies[plane_index], grids=new_grid,
                                     compute_deflections=compute_deflections))

    def generate_image_of_galaxy_light_profiles(self):
        """Generate the image of the galaxies over the entire ray trace."""
        return sum(np.array(list(map(lambda plane: plane.generate_image_of_galaxy_light_profiles(),
                                     self.planes))))

    def generate_blurring_image_of_galaxy_light_profiles(self):
        """Generate the image of all galaxy light profiles in the blurring regions of the image."""
        return np.ndarray.sum(np.array(list(map(lambda plane: plane.blurring_image_from_galaxy_light_profiles(),
                                                self.planes))))

    def reconstructors_from_planes(self, borders, cluster_mask):
        return list(map(lambda plane: plane.reconstructor_from_plane(borders, cluster_mask), self.planes))


class TracerGeometry(object):

    def __init__(self, redshifts, cosmology):
        """The geometry of a ray-tracing grid comprising an image-plane and source-plane.

        This sets up the angular diameter distances between each plane and the Earth, and between one another. \
        The critical density of the lens plane is also computed.

        Parameters
        ----------
        cosmology : astropy.cosmology.Planck15
            The cosmology of the ray-tracing calculation.
        """
        self.cosmology = cosmology
        self.redshifts = redshifts
        self.final_plane = len(self.redshifts) - 1
        self.ang_to_final_plane = self.ang_to_earth(plane_i=self.final_plane)

    def arcsec_per_kpc(self, plane_i):
        return self.cosmology.arcsec_per_kpc_proper(z=self.redshifts[plane_i]).value

    def kpc_per_arcsec(self, plane_i):
        return 1.0 / self.cosmology.arcsec_per_kpc_proper(z=self.redshifts[plane_i]).value

    def ang_to_earth(self, plane_i):
        return self.cosmology.angular_diameter_distance(self.redshifts[plane_i]).to('kpc').value

    def ang_between_planes(self, plane_i, plane_j):
        return self.cosmology.angular_diameter_distance_z1z2(self.redshifts[plane_i], self.redshifts[plane_j]). \
            to('kpc').value

    @property
    def constant_kpc(self):
        # noinspection PyUnresolvedReferences
        return constants.c.to('kpc / s').value ** 2.0 / (4 * math.pi * constants.G.to('kpc3 / M_sun s2').value)

    def critical_density_kpc(self, plane_i, plane_j):
        return self.constant_kpc * self.ang_to_earth(plane_j) / \
               (self.ang_between_planes(plane_i, plane_j) * self.ang_to_earth(plane_i))

    def critical_density_arcsec(self, plane_i, plane_j):
        return self.critical_density_kpc(plane_i, plane_j) * self.kpc_per_arcsec(plane_i) ** 2.0

    def scaling_factor(self, plane_i, plane_j):
        return (self.ang_between_planes(plane_i, plane_j) * self.ang_to_final_plane) / (
                self.ang_to_earth(plane_j) * self.ang_between_planes(plane_i, self.final_plane))


class Plane(object):

    def __init__(self, galaxies, grids, compute_deflections=True):
        """

        Represents a plane, which is a set of galaxies and grids at a given redshift in the lens ray-tracing
        calculation.

        The image-plane coordinates are defined on the observed image's uniform regular grid_coords.
        Calculating its model images from its light profiles exploits this uniformity to perform more efficient and
        precise calculations via an iterative sub-griding approach.

        The light profiles of galaxies at higher redshifts (and therefore in different lens-planes) can be assigned to
        the ImagePlane. This occurs when:

        1) The efficiency and precision offered by computing the light profile on a uniform grid_coords is preferred and
        won't lead noticeable inaccuracy. For example, computing the light profile of the main lens galaxy, ignoring
        minor lensing effects due to a low mass foreground substructure.

        2) When evaluating the light profile in its lens-plane is inaccurate. For example, when modeling the
        point-source images of a lensed quasar, effects like micro-lensing mean lens-plane modeling will be inaccurate.


        Parameters ---------- galaxies : [Galaxy] The galaxies in the plane. grids :
        mask.GridCollection The grids of (x,y) coordinates in the plane, including the image grid_coords,
        sub-grid_coords, blurring, grid_coords, etc.
        """
        self.galaxies = galaxies
        self.grids = grids

        if compute_deflections:
            def calculate_deflections(grid):
                return sum(map(lambda galaxy: galaxy.deflections_from_grid(grid), galaxies))

            self.deflections = self.grids.apply_function(calculate_deflections)

    def trace_to_next_plane(self):
        """Trace the grids to the next plane.

        NOTE : This does not work for multi-plane lensing, which requires one to use the previous plane's deflection
        angles to perform the tracing. I guess we'll ultimately call this class 'LensPlanes' and have it as a list.
        """
        return self.grids.map_function(np.subtract, self.deflections)

    def generate_image_of_galaxy_light_profiles(self):
        """Generate the image of the galaxies in this plane."""
        if len(self.galaxies) == 0:
            return np.zeros(self.grids.image.shape[0])
        return intensities_via_sub_grid(self.grids.sub, self.galaxies)

    @property
    def galaxy_images(self):
        """
        Returns
        -------
        galaxy_images: [ndarray]
            A list of images of galaxies in this plane
        """
        return [self.image_from_galaxy(galaxy) for galaxy in self.galaxies]

    @property
    def hyper_galaxies(self):
        return [galaxy.hyper_galaxy for galaxy in self.galaxies]

    def image_from_galaxy(self, galaxy):
        """
        Parameters
        ----------
        galaxy: Galaxy
            An individual galaxy, assumed to be in this plane

        Returns
        -------
        galaxy_image: ndarray
            An array describing the intensity of light coming from the galaxy embedded in this plane
        """
        return intensities_via_sub_grid(self.grids.sub, [galaxy])

    def blurring_image_from_galaxy_light_profiles(self):
        """Generate the image of the galaxies in this plane."""
        return intensities_via_grid(self.grids.blurring, self.galaxies)

    def reconstructor_from_plane(self, borders, sparse_mask):

        pixelized_galaxies = list(filter(lambda galaxy: galaxy.has_pixelization, self.galaxies))

        if len(pixelized_galaxies) == 0:
            return None
        if len(pixelized_galaxies) == 1:
            return pixelized_galaxies[0].pixelization.reconstructor_from_pix_grids(self.grids, borders, sparse_mask)
        elif len(pixelized_galaxies) > 1:
            raise exc.PixelizationException('The number of galaxies with pixelizations in one plane is above 1')


def intensities_via_sub_grid(sub_grid, galaxies):
    sub_intensities = sum(map(lambda g: g.intensity_from_grid(sub_grid), galaxies))
    return sub_grid.sub_data_to_image(sub_intensities)


def intensities_via_grid(image_grid, galaxies):
    return sum(map(lambda g: g.intensity_from_grid(image_grid), galaxies))


def deflections_for_image_grid(image_grid, galaxies):
    return sum(map(lambda galaxy: galaxy.deflections_from_grid(image_grid), galaxies))


def deflections_for_grids(grids, galaxies):
    return grids.apply_function(lambda grid: deflections_for_image_grid(grid, galaxies))


def traced_collection_for_deflections(grids, deflections):
    def subtract_scaled_deflections(grid, scaled_deflection):
        return np.subtract(grid, scaled_deflection)

    result = grids.map_function(subtract_scaled_deflections, deflections)

    return result



import os
import shutil
from functools import wraps

import pytest
from autolens import config
from autolens import exc
from autolens.autopipe import model_mapper
from autolens.autopipe import non_linear
from autolens.profiles import light_profiles, mass_profiles


@pytest.fixture(name='nlo_setup_path')
def test_nlo_setup():
    nlo_setup_path = "{}/../test_files/non_linear/nlo/setup/".format(os.path.dirname(os.path.realpath(__file__)))

    print(nlo_setup_path)

    if os.path.exists(nlo_setup_path):
        shutil.rmtree(nlo_setup_path)

    os.mkdir(nlo_setup_path)

    return nlo_setup_path


@pytest.fixture(name='nlo_paramnames_path')
def test_nlo_paramnames():
    nlo_paramnames_path = "{}/../test_files/non_linear/nlo/paramnames/".format(
        os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(nlo_paramnames_path):
        shutil.rmtree(nlo_paramnames_path)

    return nlo_paramnames_path


@pytest.fixture(name='nlo_model_info_path')
def test_nlo_model_info():
    nlo_model_info_path = "{}/../test_files/non_linear/nlo/model_info/".format(
        os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(nlo_model_info_path):
        shutil.rmtree(nlo_model_info_path)

    return nlo_model_info_path


@pytest.fixture(name='nlo_wrong_info_path')
def test_nlo_wrong_info():
    nlo_wrong_info_path = "{}/../test_files/non_linear/nlo/wrong_info/".format(
        os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(nlo_wrong_info_path):
        shutil.rmtree(nlo_wrong_info_path)

    os.mkdir(nlo_wrong_info_path)

    return nlo_wrong_info_path


@pytest.fixture(name='mn_summary_path')
def test_mn_summary():
    mn_summary_path = "{}/../test_files/non_linear/multinest/summary/".format(
        os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(mn_summary_path):
        shutil.rmtree(mn_summary_path)

    os.mkdir(mn_summary_path)

    return mn_summary_path


@pytest.fixture(name='mn_priors_path')
def test_mn_priors():
    mn_priors_path = "{}/../test_files/non_linear/multinest/priors/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(mn_priors_path):
        shutil.rmtree(mn_priors_path)

    os.mkdir(mn_priors_path)

    return mn_priors_path


@pytest.fixture(name='mn_samples_path')
def test_mn_samples():
    mn_samples_path = "{}/../test_files/non_linear/multinest/samples/".format(
        os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(mn_samples_path):
        shutil.rmtree(mn_samples_path)

    os.mkdir(mn_samples_path)

    return mn_samples_path


@pytest.fixture(name='config_path')
def test_config_path():
    return "{}/../test_files/config/priors/default/".format(os.path.dirname(os.path.realpath(__file__)))


def create_path(func):
    @wraps(func)
    def wrapper(path):
        if not os.path.exists(path):
            os.makedirs(path)
        return func(path)

    return wrapper


@create_path
def create_summary_4_parameters(path):
    summary = open(path + '/summary.txt', 'w')
    summary.write('    0.100000000000000000E+01   -0.200000000000000000E+01    0.300000000000000000E+01'
                  '    0.400000000000000000E+01   -0.500000000000000000E+01    0.600000000000000000E+01'
                  '    0.700000000000000000E+01    0.800000000000000000E+01'
                  '    0.900000000000000000E+01   -1.000000000000000000E+01   -1.100000000000000000E+01'
                  '    1.200000000000000000E+01    1.300000000000000000E+01   -1.400000000000000000E+01'
                  '   -1.500000000000000000E+01    1.600000000000000000E+01'
                  '    0.020000000000000000E+00    0.999999990000000000E+07'
                  '    0.020000000000000000E+00    0.999999990000000000E+07\n')
    summary.write('    0.100000000000000000E+01   -0.200000000000000000E+01    0.300000000000000000E+01'
                  '    0.400000000000000000E+01   -0.500000000000000000E+01    0.600000000000000000E+01'
                  '    0.700000000000000000E+01    0.800000000000000000E+01'
                  '    0.900000000000000000E+01   -1.000000000000000000E+01   -1.100000000000000000E+01'
                  '    1.200000000000000000E+01    1.300000000000000000E+01   -1.400000000000000000E+01'
                  '   -1.500000000000000000E+01    1.600000000000000000E+01'
                  '    0.020000000000000000E+00    0.999999990000000000E+07')
    summary.close()


@create_path
def create_summary_10_parameters(path):
    summary = open(path + '/summary.txt', 'w')
    summary.write('    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01'
                  '    0.400000000000000000E+01   -0.500000000000000000E+01   -0.600000000000000000E+01'
                  '   -0.700000000000000000E+01   -0.800000000000000000E+01    0.900000000000000000E+01'
                  '    1.000000000000000000E+01    1.100000000000000000E+01    1.200000000000000000E+01'
                  '    1.300000000000000000E+01    1.400000000000000000E+01    1.500000000000000000E+01'
                  '    1.600000000000000000E+01   -1.700000000000000000E+01   -1.800000000000000000E+01'
                  '    1.900000000000000000E+01    2.000000000000000000E+01    2.100000000000000000E+01'
                  '    2.200000000000000000E+01    2.300000000000000000E+01    2.400000000000000000E+01'
                  '    2.500000000000000000E+01   -2.600000000000000000E+01   -2.700000000000000000E+01'
                  '    2.800000000000000000E+01    2.900000000000000000E+01    3.000000000000000000E+01'
                  '    3.100000000000000000E+01    3.200000000000000000E+01    3.300000000000000000E+01'
                  '    3.400000000000000000E+01   -3.500000000000000000E+01   -3.600000000000000000E+01'
                  '    3.700000000000000000E+01   -3.800000000000000000E+01   -3.900000000000000000E+01'
                  '    4.000000000000000000E+01'
                  '    0.020000000000000000E+00    0.999999990000000000E+07'
                  '    0.020000000000000000E+00    0.999999990000000000E+07\n')
    summary.write('    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01'
                  '    0.400000000000000000E+01   -0.500000000000000000E+01   -0.600000000000000000E+01'
                  '   -0.700000000000000000E+01   -0.800000000000000000E+01    0.900000000000000000E+01'
                  '    1.000000000000000000E+01    1.100000000000000000E+01    1.200000000000000000E+01'
                  '    1.300000000000000000E+01    1.400000000000000000E+01    1.500000000000000000E+01'
                  '    1.600000000000000000E+01   -1.700000000000000000E+01   -1.800000000000000000E+01'
                  '    1.900000000000000000E+01    2.000000000000000000E+01    2.100000000000000000E+01'
                  '    2.200000000000000000E+01    2.300000000000000000E+01    2.400000000000000000E+01'
                  '    2.500000000000000000E+01   -2.600000000000000000E+01   -2.700000000000000000E+01'
                  '    2.800000000000000000E+01    2.900000000000000000E+01    3.000000000000000000E+01'
                  '    3.100000000000000000E+01    3.200000000000000000E+01    3.300000000000000000E+01'
                  '    3.400000000000000000E+01   -3.500000000000000000E+01   -3.600000000000000000E+01'
                  '    3.700000000000000000E+01   -3.800000000000000000E+01   -3.900000000000000000E+01'
                  '    4.000000000000000000E+01'
                  '   -1.900000000000000000E+01   -2.000000000000000000E+01')
    summary.close()


@create_path
def create_gaussian_prior_summary_4_parameters(path):
    summary = open(path + '/summary.txt', 'w')
    summary.write('    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01'
                  '    0.410000000000000000E+01    0.500000000000000000E+01    0.600000000000000000E+01'
                  '    0.700000000000000000E+01    0.800000000000000000E+01'
                  '    0.900000000000000000E+01    1.000000000000000000E+01    1.100000000000000000E+01'
                  '    1.200000000000000000E+01    1.300000000000000000E+01    1.400000000000000000E+01'
                  '    1.500000000000000000E+01    1.600000000000000000E+01'
                  '    0.020000000000000000E+00    0.999999990000000000E+07'
                  '    0.020000000000000000E+00    0.999999990000000000E+07\n')
    summary.write('    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01'
                  '    0.410000000000000000E+01    0.500000000000000000E+01    0.600000000000000000E+01'
                  '    0.700000000000000000E+01    0.800000000000000000E+01'
                  '    0.900000000000000000E+01    1.000000000000000000E+01    1.100000000000000000E+01'
                  '    1.200000000000000000E+01    1.300000000000000000E+01    1.400000000000000000E+01'
                  '    1.500000000000000000E+01    1.600000000000000000E+01'
                  '    0.020000000000000000E+00    0.999999990000000000E+07')
    summary.close()


@create_path
def create_weighted_samples_4_parameters(path):
    with open(path + '/multinest.txt', 'w+') as weighted_samples:
        weighted_samples.write(
            '    0.020000000000000000E+00    0.999999990000000000E+07    0.110000000000000000E+01    0.210000000000000000E+01    0.310000000000000000E+01    0.410000000000000000E+01\n'
            '    0.020000000000000000E+00    0.999999990000000000E+07    0.090000000000000000E+01    0.190000000000000000E+01    0.290000000000000000E+01    0.390000000000000000E+01\n'
            '    0.010000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n'
            '    0.050000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n'
            '    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n'
            '    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n'
            '    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n'
            '    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n'
            '    0.200000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01\n'
            '    0.300000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01')


@create_path
def create_weighted_samples_10_parameters(path):
    weighted_samples = open(path + '/multinest.txt', 'w')
    weighted_samples.write(
        '    0.020000000000000000E+00    0.999999990000000000E+07    0.110000000000000000E+01    0.210000000000000000E+01    0.310000000000000000E+01    0.410000000000000000E+01   -0.510000000000000000E+01   -0.610000000000000000E+01   -0.710000000000000000E+01   -0.810000000000000000E+01    0.910000000000000000E+01    1.010000000000000000E+01\n'
        '    0.020000000000000000E+00    0.999999990000000000E+07    0.090000000000000000E+01    0.190000000000000000E+01    0.290000000000000000E+01    0.390000000000000000E+01   -0.490000000000000000E+01   -0.590000000000000000E+01   -0.690000000000000000E+01   -0.790000000000000000E+01    0.890000000000000000E+01    0.990000000000000000E+01\n'
        '    0.010000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01   -0.500000000000000000E+01   -0.600000000000000000E+01   -0.700000000000000000E+01   -0.800000000000000000E+01    0.900000000000000000E+01    1.000000000000000000E+01\n'
        '    0.050000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01   -0.500000000000000000E+01   -0.600000000000000000E+01   -0.700000000000000000E+01   -0.800000000000000000E+01    0.900000000000000000E+01    1.000000000000000000E+01\n'
        '    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01   -0.500000000000000000E+01   -0.600000000000000000E+01   -0.700000000000000000E+01   -0.800000000000000000E+01    0.900000000000000000E+01    1.000000000000000000E+01\n'
        '    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01   -0.500000000000000000E+01   -0.600000000000000000E+01   -0.700000000000000000E+01   -0.800000000000000000E+01    0.900000000000000000E+01    1.000000000000000000E+01\n'
        '    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01   -0.500000000000000000E+01   -0.600000000000000000E+01   -0.700000000000000000E+01   -0.800000000000000000E+01    0.900000000000000000E+01    1.000000000000000000E+01\n'
        '    0.100000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01   -0.500000000000000000E+01   -0.600000000000000000E+01   -0.700000000000000000E+01   -0.800000000000000000E+01    0.900000000000000000E+01    1.000000000000000000E+01\n'
        '    0.200000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01   -0.500000000000000000E+01   -0.600000000000000000E+01   -0.700000000000000000E+01   -0.800000000000000000E+01    0.900000000000000000E+01    1.000000000000000000E+01\n'
        '    0.300000000000000000E+00    0.999999990000000000E+07    0.100000000000000000E+01    0.200000000000000000E+01    0.300000000000000000E+01    0.400000000000000000E+01   -0.500000000000000000E+01   -0.600000000000000000E+01   -0.700000000000000000E+01   -0.800000000000000000E+01    0.900000000000000000E+01    1.000000000000000000E+01')
    weighted_samples.close()


class MockClassNLOx4(object):

    component_number = 0

    def __init__(self, one=1, two=2, three=3, four=4):
        self.one = one
        self.two = two
        self.three = three
        self.four = four

    @property
    def subscript(self):
        return 'a'

    @property
    def parameter_labels(self):
        return [r'x4p0', r'x4p1', r'x4p2', r'x4p3']


class MockClassNLOx5(object):

    component_number = 0

    def __init__(self, one=1, two=2, three=3, four=4, five=5):
        self.one = one
        self.two = two
        self.three = three
        self.four = four
        self.five = five

    @property
    def subscript(self):
        return 'a'

    @property
    def parameter_labels(self):
        return [r'x4p0', r'x4p1', r'x4p2', r'x4p3']


class MockClassNLOx6(object):

    component_number = 1

    def __init__(self, one=(1,2), two=(3,4), three=3, four=4):

        self.one = one
        self.two = two
        self.three = three
        self.four = four

    @property
    def subscript(self):
        return 'b'

    @property
    def parameter_labels(self):
        return [r'x6p0', r'x6p1', r'x6p2', r'x6p3', r'x6p4', r'x6p5']


class TestGenerateLatex(object):

    def test__one_parameter__no_subscript(self):
        assert non_linear.generate_parameter_latex('x') == ['$x$']

    def test__three_parameters__no_subscript(self):
        assert non_linear.generate_parameter_latex(['x', 'y', 'z']) == ['$x$', '$y$', '$z$']

    def test__one_parameter__subscript__no_number(self):
        assert non_linear.generate_parameter_latex(['x'], subscript='d') == [r'$x_{\mathrm{d}}$']

    def test__three_parameters__subscript__no_number(self):
        assert non_linear.generate_parameter_latex(['x', 'y', 'z'], subscript='d') == [r'$x_{\mathrm{d}}$',
                                                                                       r'$y_{\mathrm{d}}$',
                                                                                       r'$z_{\mathrm{d}}$']


class TestNonLinearOptimizer(object):

    class TestDirectorySetup(object):

        def test__1_class__correct_directory(self, config_path, nlo_setup_path):

            nlo = non_linear.NonLinearOptimizer(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=nlo_setup_path + '1_class', config_path=config_path)
            nlo.variable.add_classes(parallel=MockClassNLOx4)
            nlo.save_model_info()

            assert os.path.exists(nlo_setup_path + '1_class') == True

    class TestTotalParameters:

        def test__1_class__four_parameters(self, config_path, nlo_setup_path):
            nlo = non_linear.NonLinearOptimizer(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=nlo_setup_path + '1_class', config_path=config_path)
            nlo.variable.add_classes(class_1=MockClassNLOx4)

            assert nlo.variable.total_parameters == 4

        def test__2_classes__six_parameters(self, config_path, nlo_setup_path):
            nlo = non_linear.NonLinearOptimizer(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=nlo_setup_path + '2_classes', config_path=config_path)

            nlo.variable.add_classes(class_1=MockClassNLOx4, class_2=MockClassNLOx6)

            assert nlo.variable.total_parameters == 10
    #
    # class TestCreateParamNames(object):
    #
    #     def test__1_class_and_parameter_set(self, config_path, nlo_paramnames_path):
    #
    #         nlo = non_linear.NonLinearOptimizer(model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
    #                                             path=nlo_paramnames_path, config_path=config_path)
    #
    #         nlo.variable.add_classes(mock_class=MockClassNLOx4)
    #         nlo.save_model_info()
    #
    #         paramnames_file = open(nlo_paramnames_path + '/multinest.paramnames')
    #
    #         paramnames = paramnames_file.readlines()
    #
    #         assert paramnames[0] == r'mock_class_one                          $x4p0_{\mathrm{a1}}$' + '\n'
    #         assert paramnames[1] == r'mock_class_two                          $x4p1_{\mathrm{a1}}$' + '\n'
    #         assert paramnames[2] == r'mock_class_three                        $x4p2_{\mathrm{a1}}$' + '\n'
    #         assert paramnames[3] == r'mock_class_four                         $x4p3_{\mathrm{a1}}$' + '\n'
    #
    #     def test__2_classes__includes_a_tuple(self, config_path, nlo_paramnames_path):
    #
    #         nlo = non_linear.NonLinearOptimizer(model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
    #                                             path=nlo_paramnames_path, config_path=config_path)
    #
    #         nlo.variable.add_classes(mock_class_1=MockClassNLOx4)
    #         nlo.variable.add_classes(mock_class_2=MockClassNLOx6)
    #         nlo.save_model_info()
    #
    #         paramnames_file = open(nlo_paramnames_path + '/multinest.paramnames')
    #
    #         paramnames = paramnames_file.readlines()
    #
    #         assert paramnames[0] == r'mock_class_1_one                        $x4p0_{\mathrm{a1}}$' + '\n'
    #         assert paramnames[1] == r'mock_class_1_two                        $x4p1_{\mathrm{a1}}$' + '\n'
    #         assert paramnames[2] == r'mock_class_1_three                      $x4p2_{\mathrm{a1}}$' + '\n'
    #         assert paramnames[3] == r'mock_class_1_four                       $x4p3_{\mathrm{a1}}$' + '\n'
    #         assert paramnames[4] == r'mock_class_2_one_0                      $x6p0_{\mathrm{b2}}$' + '\n'
    #         assert paramnames[5] == r'mock_class_2_one_1                      $x6p1_{\mathrm{b2}}$' + '\n'
    #         assert paramnames[6] == r'mock_class_2_two_0                      $x6p2_{\mathrm{b2}}$' + '\n'
    #         assert paramnames[7] == r'mock_class_2_two_1                      $x6p3_{\mathrm{b2}}$' + '\n'
    #         assert paramnames[8] == r'mock_class_2_three                      $x6p4_{\mathrm{b2}}$' + '\n'
    #         assert paramnames[9] == r'mock_class_2_four                       $x6p5_{\mathrm{b2}}$' + '\n'


    class TestMakeModelInfo(object):

        def test__single_class__outputs_all_info(self, config_path, nlo_model_info_path):

            nlo = non_linear.NonLinearOptimizer(model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                                                path=nlo_model_info_path, config_path=config_path)

            nlo.variable.add_classes(mock_class=MockClassNLOx4)
            nlo.save_model_info()

            model_info_file = open(nlo_model_info_path + '/model.info')

            model_info = model_info_file.readlines()

            assert model_info[0] == r'MockClassNLOx4' + '\n'
            assert model_info[1] == r'' + '\n'
            assert model_info[2] == r'one: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[3] == r'two: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[4] == r'three: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[5] == r'four: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'

        def test__two_models_and_parameter_sets__output_all_info(self, config_path, nlo_model_info_path):

            nlo = non_linear.NonLinearOptimizer(model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                                                path=nlo_model_info_path, config_path=config_path)

            nlo.variable.add_classes(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)
            nlo.save_model_info()

            model_info_file = open(nlo_model_info_path + '/model.info')

            model_info = model_info_file.readlines()

            assert model_info[0] == r'MockClassNLOx4' + '\n'
            assert model_info[1] == r'' + '\n'
            assert model_info[2] == r'one: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[3] == r'two: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[4] == r'three: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[5] == r'four: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[6] == r'' + '\n'
            assert model_info[7] == r'MockClassNLOx6' + '\n'
            assert model_info[8] == r'' + '\n'
            assert model_info[9] == r'one_0: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[10] == r'one_1: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[11] == r'two_0: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[12] == r'two_1: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[13] == r'three: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
            assert model_info[14] == r'four: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'

    class TestWrongModelInfo(object):

        def test__single_model__prior_changed_from_input_model__raises_error(self, config_path, nlo_wrong_info_path):
            with open(nlo_wrong_info_path + '/model.info', 'w') as file: file.write('The model info is missing :(')

            with pytest.raises(exc.PriorException):
                nl = non_linear.NonLinearOptimizer(
                    model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)), path=nlo_wrong_info_path)
                nl.variable.add_classes(mass_profile=mass_profiles.SphericalNFW)
                nl.save_model_info()


class TestMultiNest(object):

    class TestReadFromSummary:

        def test__read_most_probable__1_class_4_params(self, config_path, mn_summary_path):

            create_summary_4_parameters(path=mn_summary_path + '1_class')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_summary_path + '1_class')
            mn.variable.add_classes(mock_class=MockClassNLOx4)

            most_probable = mn.most_probable_from_summary()

            assert most_probable == [1.0, -2.0, 3.0, 4.0]

        def test__read_most_probable__2_classes_6_params(self, config_path, mn_summary_path):
            create_summary_10_parameters(path=mn_summary_path + '2_classes')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_summary_path + '2_classes')

            mn.variable.add_classes(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

            most_probable = mn.most_probable_from_summary()

            assert most_probable == [1.0, 2.0, 3.0, 4.0, -5.0, -6.0, -7.0, -8.0, 9.0, 10.0]

        def test__most_probable__setup_model_instance__1_class_4_params(self, config_path, mn_summary_path):
            create_summary_4_parameters(path=mn_summary_path + '1_class')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_summary_path + '/1_class')
            mn.variable.add_classes(mock_class=MockClassNLOx4)

            most_probable = mn.most_probable_instance_from_summary()

            assert most_probable.mock_class.one == 1.0
            assert most_probable.mock_class.two == -2.0
            assert most_probable.mock_class.three == 3.0
            assert most_probable.mock_class.four == 4.0

        def test__most_probable__setup_model_instance__2_classes_6_params(self, config_path, mn_summary_path):
            create_summary_10_parameters(path=mn_summary_path + '2_classes')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_summary_path + '2_classes')

            mn.variable.add_classes(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

            most_probable = mn.most_probable_instance_from_summary()

            assert most_probable.mock_class_1.one == 1.0
            assert most_probable.mock_class_1.two == 2.0
            assert most_probable.mock_class_1.three == 3.0
            assert most_probable.mock_class_1.four == 4.0

            assert most_probable.mock_class_2.one == (-5.0, -6.0)
            assert most_probable.mock_class_2.two == (-7.0, -8.0)
            assert most_probable.mock_class_2.three == 9.0
            assert most_probable.mock_class_2.four == 10.0

        def test__most_probable__setup_model_instance__1_class_5_params_but_1_is_constant(self, config_path,
                                                                                          mn_summary_path):

            create_summary_4_parameters(path=mn_summary_path + '1_class')

            mapper = model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path), mock_class=MockClassNLOx5)
            mapper.mock_class.five = model_mapper.Constant(10.0)

            mn = non_linear.MultiNest(model_mapper=mapper, path=mn_summary_path + '/1_class')

            most_probable = mn.most_probable_instance_from_summary()

            assert most_probable.mock_class.one == 1.0
            assert most_probable.mock_class.two == -2.0
            assert most_probable.mock_class.three == 3.0
            assert most_probable.mock_class.four == 4.0
            assert most_probable.mock_class.five == 10.0

        def test__read_most_likely__1_class_4_params(self, config_path, mn_summary_path):
            create_summary_4_parameters(path=mn_summary_path + '1_class')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_summary_path + '1_class')
            mn.variable.add_classes(mock_class=MockClassNLOx4)

            most_likely = mn.most_likely_from_summary()

            assert most_likely == [9.0, -10.0, -11.0, 12.0]

        def test__read_most_likely__2_classes_6_params(self, config_path, mn_summary_path):
            create_summary_10_parameters(path=mn_summary_path + '2_classes')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_summary_path + '2_classes')

            mn.variable.add_classes(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

            most_likely = mn.most_likely_from_summary()

            assert most_likely == [21.0, 22.0, 23.0, 24.0, 25.0, -26.0, -27.0, 28.0, 29.0, 30.0]

        def test__most_likely__setup_model_instance__1_class_4_params(self, config_path, mn_summary_path):
            create_summary_4_parameters(path=mn_summary_path + '1_class')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_summary_path + '1_class')
            mn.variable.add_classes(mock_class=MockClassNLOx4)

            most_likely = mn.most_likely_instance_from_summary()

            assert most_likely.mock_class.one == 9.0
            assert most_likely.mock_class.two == -10.0
            assert most_likely.mock_class.three == -11.0
            assert most_likely.mock_class.four == 12.0

        def test__most_likely__setup_model_instance__2_classes_6_params(self, config_path, mn_summary_path):
            create_summary_10_parameters(path=mn_summary_path + '2_classes')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_summary_path + '2_classes')

            mn.variable.add_classes(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

            most_likely = mn.most_likely_instance_from_summary()

            assert most_likely.mock_class_1.one == 21.0
            assert most_likely.mock_class_1.two == 22.0
            assert most_likely.mock_class_1.three == 23.0
            assert most_likely.mock_class_1.four == 24.0

            assert most_likely.mock_class_2.one == (25.0, -26.0)
            assert most_likely.mock_class_2.two == (-27.0, 28.0)
            assert most_likely.mock_class_2.three == 29.0
            assert most_likely.mock_class_2.four == 30.0

        def test__most_likely__setup_model_instance__1_class_5_params_but_1_is_constant(self, config_path,
                                                                                          mn_summary_path):

            create_summary_4_parameters(path=mn_summary_path + '1_class')

            mapper = model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path), mock_class=MockClassNLOx5)
            mapper.mock_class.five = model_mapper.Constant(10.0)

            mn = non_linear.MultiNest(model_mapper=mapper, path=mn_summary_path + '/1_class')

            most_likely = mn.most_likely_instance_from_summary()

            assert most_likely.mock_class.one == 9.0
            assert most_likely.mock_class.two == -10.0
            assert most_likely.mock_class.three == -11.0
            assert most_likely.mock_class.four == 12.0
            assert most_likely.mock_class.five == 10.0

    class TestGaussianPriors(object):

        def test__1_class__gaussian_priors_at_3_sigma_confidence(self, config_path, mn_priors_path):
            create_gaussian_prior_summary_4_parameters(path=mn_priors_path)
            create_weighted_samples_4_parameters(path=mn_priors_path)

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_priors_path)

            mn.variable.add_classes(mock_class=MockClassNLOx4)

            gaussian_priors = mn.gaussian_priors_at_sigma_limit(sigma_limit=3.0)

            assert gaussian_priors[0][0] == 1.0
            assert gaussian_priors[1][0] == 2.0
            assert gaussian_priors[2][0] == 3.0
            assert gaussian_priors[3][0] == 4.1

            assert gaussian_priors[0][1] == pytest.approx(0.12, 1e-2)
            assert gaussian_priors[1][1] == pytest.approx(0.12, 1e-2)
            assert gaussian_priors[2][1] == pytest.approx(0.12, 1e-2)
            assert gaussian_priors[3][1] == pytest.approx(0.22, 1e-2)

        def test__1_profile__gaussian_priors_at_1_sigma_confidence(self, config_path, mn_priors_path):
            create_gaussian_prior_summary_4_parameters(path=mn_priors_path)
            create_weighted_samples_4_parameters(path=mn_priors_path)

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_priors_path)

            mn.variable.add_classes(mock_class=MockClassNLOx4)
            gaussian_priors = mn.gaussian_priors_at_sigma_limit(sigma_limit=1.0)

            # Use sigmas directly as rouding errors come in otherwise
            lower_sigmas = mn.model_at_lower_sigma_limit(sigma_limit=1.0)

            assert gaussian_priors[0][0] == 1.0
            assert gaussian_priors[1][0] == 2.0
            assert gaussian_priors[2][0] == 3.0
            assert gaussian_priors[3][0] == 4.1

            assert gaussian_priors[0][1] == pytest.approx(1.0 - lower_sigmas[0], 5e-2)
            assert gaussian_priors[1][1] == pytest.approx(2.0 - lower_sigmas[1], 5e-2)
            assert gaussian_priors[2][1] == pytest.approx(3.0 - lower_sigmas[2], 5e-2)
            assert gaussian_priors[3][1] == pytest.approx(4.1 - lower_sigmas[3], 5e-2)

    class TestWeightedSamples(object):

        def test__1_class__1st_first_weighted_sample__model_weight_and_likelihood(self, config_path, mn_samples_path):
            create_weighted_samples_4_parameters(path=mn_samples_path + '1_class')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_samples_path + '1_class')

            mn.variable.add_classes(mock_class=MockClassNLOx4)

            model, weight, likelihood = mn.weighted_sample_model_from_weighted_samples(index=0)

            assert model == [1.1, 2.1, 3.1, 4.1]
            assert weight == 0.02
            assert likelihood == -0.5 * 9999999.9

        def test__1_class__5th_weighted_sample__model_weight_and_likelihood(self, config_path, mn_samples_path):
            create_weighted_samples_4_parameters(path=mn_samples_path + '1_class')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_samples_path + '1_class')

            mn.variable.add_classes(mock_class=MockClassNLOx4)

            model, weight, likelihood = mn.weighted_sample_model_from_weighted_samples(index=5)

            assert model == [1.0, 2.0, 3.0, 4.0]
            assert weight == 0.1
            assert likelihood == -0.5 * 9999999.9

        def test__2_classes__1st_weighted_sample__model_weight_and_likelihood(self, config_path, mn_samples_path):
            create_weighted_samples_10_parameters(path=mn_samples_path + '2_classes')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_samples_path + '2_classes')

            mn.variable.add_classes(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

            model, weight, likelihood = mn.weighted_sample_model_from_weighted_samples(index=0)

            assert model == [1.1, 2.1, 3.1, 4.1, -5.1, -6.1, -7.1, -8.1, 9.1, 10.1]
            assert weight == 0.02
            assert likelihood == -0.5 * 9999999.9

        def test__2_classes__5th_weighted_sample__model_weight_and_likelihood(self, config_path,
                                                                              mn_samples_path):
            create_weighted_samples_10_parameters(path=mn_samples_path + '1_classes')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_samples_path + '1_classes')

            mn.variable.add_classes(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

            model, weight, likelihood = mn.weighted_sample_model_from_weighted_samples(index=5)

            assert model == [1.0, 2.0, 3.0, 4.0, -5.0, -6.0, -7.0, -8.0, 9.0, 10.0]
            assert weight == 0.1
            assert likelihood == -0.5 * 9999999.9

        def test__1_class__1st_weighted_sample_model_instance__include_weight_and_likelihood(self, config_path,
                                                                                             mn_samples_path):
            create_weighted_samples_4_parameters(path=mn_samples_path + '1_class')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_samples_path + '1_class')

            mn.variable.add_classes(mock_class=MockClassNLOx4)

            weighted_sample_model, weight, likelihood = mn.weighted_sample_instance_from_weighted_samples(index=0)

            assert weight == 0.02
            assert likelihood == -0.5 * 9999999.9

            assert weighted_sample_model.mock_class.one == 1.1
            assert weighted_sample_model.mock_class.two == 2.1
            assert weighted_sample_model.mock_class.three == 3.1
            assert weighted_sample_model.mock_class.four == 4.1

        def test__1_class__5th_weighted_sample_model_instance__include_weight_and_likelihood(self, config_path,
                                                                                             mn_samples_path):
            create_weighted_samples_4_parameters(path=mn_samples_path + '1_class')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_samples_path + '1_class')

            mn.variable.add_classes(mock_class=MockClassNLOx4)

            weighted_sample_model, weight, likelihood = mn.weighted_sample_instance_from_weighted_samples(index=5)

            assert weight == 0.1
            assert likelihood == -0.5 * 9999999.9

            assert weighted_sample_model.mock_class.one == 1.0
            assert weighted_sample_model.mock_class.two == 2.0
            assert weighted_sample_model.mock_class.three == 3.0
            assert weighted_sample_model.mock_class.four == 4.0

        def test__2_classes__1st_weighted_sample_model_instance__include_weight_and_likelihood(self, config_path,
                                                                                               mn_samples_path):
            create_weighted_samples_10_parameters(path=mn_samples_path + '1_classes')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_samples_path + '1_classes')

            mn.variable.add_classes(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

            weighted_sample_model, weight, likelihood = mn.weighted_sample_instance_from_weighted_samples(index=0)

            assert weight == 0.02
            assert likelihood == -0.5 * 9999999.9

            assert weighted_sample_model.mock_class_1.one == 1.1
            assert weighted_sample_model.mock_class_1.two == 2.1
            assert weighted_sample_model.mock_class_1.three == 3.1
            assert weighted_sample_model.mock_class_1.four == 4.1

            assert weighted_sample_model.mock_class_2.one == (-5.1, -6.1)
            assert weighted_sample_model.mock_class_2.two == (-7.1, -8.1)
            assert weighted_sample_model.mock_class_2.three == 9.1
            assert weighted_sample_model.mock_class_2.four == 10.1

        def test__2_classes__5th_weighted_sample_model_instance__include_weight_and_likelihood(self, config_path,
                                                                                               mn_samples_path):
            create_weighted_samples_10_parameters(path=mn_samples_path + '1_classes')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_samples_path + '1_classes')

            mn.variable.add_classes(mock_class_1=MockClassNLOx4, mock_class_2=MockClassNLOx6)

            weighted_sample_model, weight, likelihood = mn.weighted_sample_instance_from_weighted_samples(index=5)

            assert weight == 0.1
            assert likelihood == -0.5 * 9999999.9

            assert weighted_sample_model.mock_class_1.one == 1.0
            assert weighted_sample_model.mock_class_1.two == 2.0
            assert weighted_sample_model.mock_class_1.three == 3.0
            assert weighted_sample_model.mock_class_1.four == 4.0

            assert weighted_sample_model.mock_class_2.one == (-5.0, -6.0)
            assert weighted_sample_model.mock_class_2.two == (-7.0, -8.0)
            assert weighted_sample_model.mock_class_2.three == 9.0
            assert weighted_sample_model.mock_class_2.four == 10.0

    class TestLimits(object):

        def test__1_profile__limits_1d_vectors_via_weighted_samples__1d_vectors_are_correct(self, config_path,
                                                                                            mn_samples_path):
            create_weighted_samples_4_parameters(path=mn_samples_path + '1_class')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_samples_path + '1_class')

            mn.variable.add_classes(mock_class=MockClassNLOx4)

            assert mn.model_at_upper_sigma_limit(sigma_limit=3.0) == pytest.approx([1.12, 2.12, 3.12, 4.12], 1e-2)
            assert mn.model_at_lower_sigma_limit(sigma_limit=3.0) == pytest.approx([0.88, 1.88, 2.88, 3.88], 1e-2)

        def test__1_profile__change_limit_to_1_sigma(self, config_path, mn_samples_path):
            create_weighted_samples_4_parameters(path=mn_samples_path + '1_class')

            mn = non_linear.MultiNest(
                model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
                path=mn_samples_path + '1_class')

            mn.variable.add_classes(mock_class=MockClassNLOx4)

            assert mn.model_at_upper_sigma_limit(sigma_limit=1.0) == pytest.approx([1.07, 2.07, 3.07, 4.07], 1e-2)
            assert mn.model_at_lower_sigma_limit(sigma_limit=1.0) == pytest.approx([0.93, 1.93, 2.93, 3.93], 1e-2)


class TestConfig(object):
    def test_multinest_default(self):
        multinest = non_linear.MultiNest()

        assert multinest.importance_nested_sampling is True

        assert multinest.multimodal is True
        assert multinest.const_efficiency_mode is False
        assert multinest.n_live_points == 400

        assert multinest.evidence_tolerance == 0.5
        assert multinest.sampling_efficiency == 0.8

        assert multinest.n_iter_before_update == 100
        assert multinest.null_log_evidence == -1e90

        assert multinest.max_modes == 100
        assert multinest.mode_tolerance == -1e90

        assert multinest.outputfiles_basename == "chains/1-"
        assert multinest.seed == -1
        assert multinest.verbose is False

        assert multinest.resume is True
        assert multinest.context == 0
        assert multinest.write_output is True
        assert multinest.log_zero == -1e100

        assert multinest.max_iter == 0
        assert multinest.init_MPI is False

    def test_downhill_simplex_default(self):
        downhill_simplex = non_linear.DownhillSimplex()

        assert downhill_simplex.xtol == 1e-4
        assert downhill_simplex.ftol == 1e-4
        assert downhill_simplex.maxiter is None
        assert downhill_simplex.maxfun is None

        assert downhill_simplex.full_output == 0
        assert downhill_simplex.disp == 1
        assert downhill_simplex.retall == 0


class TestRealClasses(object):

    def test__directory_setup__input_path_sets_up(self, config_path, nlo_setup_path):

        nlo = non_linear.NonLinearOptimizer(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=nlo_setup_path + '1_profile')

        nlo.variable.add_classes(light_profile=light_profiles.EllipticalSersic)
        nlo.save_model_info()

        assert os.path.exists(nlo_setup_path + '1_profile') == True

    def test__number_of_params__multiple_light_and_mass_profiles(self, config_path, nlo_setup_path):

        nlo = non_linear.NonLinearOptimizer(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=nlo_setup_path)

        nlo.variable.add_classes(light_profile=light_profiles.EllipticalSersic,
                                 light_profile_2=light_profiles.EllipticalSersic,
                                 light_profile_3=light_profiles.EllipticalSersic,
                                 mass_profile=mass_profiles.SphericalNFW,
                                 mass_profile_2=mass_profiles.SphericalNFW)

        assert nlo.variable.total_parameters == 29

    # def test__create_param_names_2_light_models__2_mass_models(self, config_path, nlo_paramnames_path):
    #
    #     light_profiles.EllipticalLightProfile._ids = count()
    #     mass_profiles.EllipticalMassProfile._ids = count()
    #
    #     nlo = non_linear.NonLinearOptimizer(
    #         model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
    #         path=nlo_paramnames_path)
    #
    #     nlo.variable.add_classes(
    #         light_profile_0=light_profiles.EllipticalSersic,
    #         light_profile_1=light_profiles.EllipticalExponential,
    #         mass_profile_0=mass_profiles.SphericalIsothermal,
    #         mass_profile_1=mass_profiles.SphericalNFW)
    #
    #     nlo.save_model_info()
    #
    #     paramnames_test = open(nlo_paramnames_path + 'multinest.paramnames')
    #
    #     paramnames_str = paramnames_test.readlines()
    #
    #     assert paramnames_str[0] == r'light_profile_0_centre_0                $x_{\mathrm{l1}}$' + '\n'
    #     assert paramnames_str[1] == r'light_profile_0_centre_1                $y_{\mathrm{l1}}$' + '\n'
    #     assert paramnames_str[2] == r'light_profile_0_axis_ratio              $q_{\mathrm{l1}}$' + '\n'
    #     assert paramnames_str[3] == r'light_profile_0_phi                     $\phi_{\mathrm{l1}}$' + '\n'
    #     assert paramnames_str[4] == r'light_profile_0_intensity               $I_{\mathrm{l1}}$' + '\n'
    #     assert paramnames_str[5] == r'light_profile_0_effective_radius        $R_{\mathrm{l1}}$' + '\n'
    #     assert paramnames_str[6] == r'light_profile_0_sersic_index            $n_{\mathrm{l1}}$' + '\n'
    #     assert paramnames_str[7] == r'light_profile_1_centre_0                $x_{\mathrm{l2}}$' + '\n'
    #     assert paramnames_str[8] == r'light_profile_1_centre_1                $y_{\mathrm{l2}}$' + '\n'
    #     assert paramnames_str[9] == r'light_profile_1_axis_ratio              $q_{\mathrm{l2}}$' + '\n'
    #     assert paramnames_str[10] == r'light_profile_1_phi                     $\phi_{\mathrm{l2}}$' + '\n'
    #     assert paramnames_str[11] == r'light_profile_1_intensity               $I_{\mathrm{l2}}$' + '\n'
    #     assert paramnames_str[12] == r'light_profile_1_effective_radius        $R_{\mathrm{l2}}$' + '\n'
    #     assert paramnames_str[13] == r'mass_profile_0_centre_0                 $x_{\mathrm{1}}$' + '\n'
    #     assert paramnames_str[14] == r'mass_profile_0_centre_1                 $y_{\mathrm{1}}$' + '\n'
    #     assert paramnames_str[15] == r'mass_profile_0_einstein_radius          $\theta_{\mathrm{1}}$' + '\n'
    #     assert paramnames_str[16] == r'mass_profile_1_centre_0                 $x_{\mathrm{d2}}$' + '\n'
    #     assert paramnames_str[17] == r'mass_profile_1_centre_1                 $y_{\mathrm{d2}}$' + '\n'
    #     assert paramnames_str[18] == r'mass_profile_1_kappa_s                  $\kappa_{\mathrm{d2}}$' + '\n'
    #     assert paramnames_str[19] == r'mass_profile_1_scale_radius             $Rs_{\mathrm{d2}}$' + '\n'

    def test__output_model_info__2_models(self, config_path, nlo_model_info_path):
        nlo = non_linear.NonLinearOptimizer(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=nlo_model_info_path)

        nlo.variable.add_classes(light_profile_0=light_profiles.EllipticalSersic,
                                 light_profile_1=light_profiles.EllipticalExponential,
                                 mass_profile_0=mass_profiles.SphericalIsothermal,
                                 mass_profile_1=mass_profiles.SphericalNFW)

        nlo.save_model_info()

        model_info_test = open(nlo_model_info_path + 'model.info')

        model_info_str = model_info_test.readlines()

        assert model_info_str[0] == r'EllipticalSersic' + '\n'
        assert model_info_str[1] == r'' + '\n'
        assert model_info_str[2] == r'centre_0: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
        assert model_info_str[3] == r'centre_1: UniformPrior, lower_limit = 0.0, upper_limit = 0.5' + '\n'
        assert model_info_str[4] == r'axis_ratio: UniformPrior, lower_limit = 0.0, upper_limit = 0.5' + '\n'
        assert model_info_str[5] == r'phi: UniformPrior, lower_limit = 0.0, upper_limit = 0.5' + '\n'
        assert model_info_str[6] == r'intensity: GaussianPrior, mean = 0.0, sigma = 0.5' + '\n'
        assert model_info_str[7] == r'effective_radius: UniformPrior, lower_limit = 1.0, upper_limit = 1.0' + '\n'
        assert model_info_str[8] == r'sersic_index: UniformPrior, lower_limit = 1.0, upper_limit = 1.0' + '\n'
        assert model_info_str[9] == r'' + '\n'
        assert model_info_str[10] == r'EllipticalExponential' + '\n'
        assert model_info_str[11] == r'' + '\n'
        assert model_info_str[12] == r'centre_0: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
        assert model_info_str[13] == r'centre_1: UniformPrior, lower_limit = 0.0, upper_limit = 0.5' + '\n'
        assert model_info_str[14] == r'axis_ratio: UniformPrior, lower_limit = 0.0, upper_limit = 0.5' + '\n'
        assert model_info_str[15] == r'phi: UniformPrior, lower_limit = 0.0, upper_limit = 0.5' + '\n'
        assert model_info_str[16] == r'intensity: GaussianPrior, mean = 0.0, sigma = 0.5' + '\n'
        assert model_info_str[17] == r'effective_radius: UniformPrior, lower_limit = 1.0, upper_limit = 1.0' + '\n'
        assert model_info_str[18] == r'' + '\n'
        assert model_info_str[19] == r'SphericalIsothermal' + '\n'
        assert model_info_str[20] == r'' + '\n'
        assert model_info_str[21] == r'centre_0: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
        assert model_info_str[22] == r'centre_1: UniformPrior, lower_limit = 0.0, upper_limit = 0.5' + '\n'
        assert model_info_str[23] == r'einstein_radius: UniformPrior, lower_limit = 1.0, upper_limit = 1.0' + '\n'
        assert model_info_str[24] == r'' + '\n'
        assert model_info_str[25] == r'SphericalNFW' + '\n'
        assert model_info_str[26] == r'' + '\n'
        assert model_info_str[27] == r'centre_0: UniformPrior, lower_limit = 0.0, upper_limit = 1.0' + '\n'
        assert model_info_str[28] == r'centre_1: UniformPrior, lower_limit = 0.0, upper_limit = 0.5' + '\n'
        assert model_info_str[29] == r'kappa_s: UniformPrior, lower_limit = 1.0, upper_limit = 1.0' + '\n'
        assert model_info_str[30] == r'scale_radius: UniformPrior, lower_limit = 1.0, upper_limit = 1.0' + '\n'

    def test__read_multinest_most_probable_via_summary__multiple_profiles(self, config_path, mn_summary_path):
        create_summary_10_parameters(path=mn_summary_path + 'multi_profile')

        files = non_linear.MultiNest(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=mn_summary_path + 'multi_profile')

        files.variable.add_classes(light_profile=light_profiles.EllipticalExponential,
                                   mass_profile=mass_profiles.SphericalNFW)

        files.save_model_info()

        most_probable = files.most_probable_from_summary()

        assert most_probable == [1.0, 2.0, 3.0, 4.0, -5.0, -6.0, -7.0, -8.0, 9.0, 10.0]

    def test__setup_most_probable_model_instance_via_multinest_summary(self, config_path, mn_summary_path):
        create_summary_10_parameters(path=mn_summary_path + 'multi_profile')

        files = non_linear.MultiNest(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=mn_summary_path + 'multi_profile')

        files.variable.add_classes(light_profile=light_profiles.EllipticalExponential,
                                   mass_profile=mass_profiles.SphericalNFW)

        files.save_model_info()

        most_likely = files.most_likely_from_summary()

        assert most_likely == [21.0, 22.0, 23.0, 24.0, 25.0, -26.0, -27.0, 28.0, 29.0, 30.0]

    def test__read_multinest_most_likely_via_summary__multiple_profiles(self, config_path, mn_summary_path):
        create_summary_10_parameters(path=mn_summary_path + 'multi_profile')

        files = non_linear.MultiNest(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=mn_summary_path + 'multi_profile')

        files.variable.add_classes(light_profile=light_profiles.EllipticalExponential,
                                   mass_profile=mass_profiles.SphericalNFW)

        files.save_model_info()

        max_likelihood = files.max_likelihood_from_summary()
        max_log_likelihood = files.max_log_likelihood_from_summary()

        assert max_likelihood == 0.02
        assert max_log_likelihood == 9999999.9

    def test__setup_most_likely_model_instance_via_multinest_summaryy(self, config_path, mn_summary_path):

        create_summary_10_parameters(path=mn_summary_path + 'multi_profile')

        multinest = non_linear.MultiNest(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=mn_summary_path + 'multi_profile')
        multinest.variable.add_classes(light_profile=light_profiles.EllipticalExponential,
                                       mass_profile=mass_profiles.SphericalNFW)

        multinest.save_model_info()

        most_probable = multinest.most_probable_instance_from_summary()
        most_likely = multinest.most_likely_instance_from_summary()

        assert most_probable.light_profile.centre == (1.0, 2.0)
        assert most_probable.light_profile.axis_ratio == 3.0
        assert most_probable.light_profile.phi == 4.0
        assert most_probable.light_profile.intensity == -5.0
        assert most_probable.light_profile.effective_radius == -6.0

        assert most_probable.mass_profile.centre == (-7.0, -8.0)
        assert most_probable.mass_profile.kappa_s == 9.0
        assert most_probable.mass_profile.scale_radius == 10.0

        assert most_likely.light_profile.centre == (21.0, 22.0)
        assert most_likely.light_profile.axis_ratio == 23.0
        assert most_likely.light_profile.phi == 24.0
        assert most_likely.light_profile.intensity == 25.0
        assert most_likely.light_profile.effective_radius == -26.0

        assert most_likely.mass_profile.centre == (-27.0, 28.0)
        assert most_likely.mass_profile.kappa_s == 29.0
        assert most_likely.mass_profile.scale_radius == 30.0

    def test__gaussian_priors_at_3_sigma_confidence__1_profile(self, config_path, mn_priors_path):
        create_gaussian_prior_summary_4_parameters(mn_priors_path)
        create_weighted_samples_4_parameters(mn_priors_path)

        results = non_linear.MultiNest(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=mn_priors_path)
        results.variable.add_classes(mass_profile=mass_profiles.SphericalNFW)

        results.save_model_info()

        gaussian_priors = results.gaussian_priors_at_sigma_limit(sigma_limit=3.0)

        assert gaussian_priors[0][0] == 1.0
        assert gaussian_priors[1][0] == 2.0
        assert gaussian_priors[2][0] == 3.0
        assert gaussian_priors[3][0] == 4.1

        assert gaussian_priors[0][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[1][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[2][1] == pytest.approx(0.12, 1e-2)
        assert gaussian_priors[3][1] == pytest.approx(0.22, 1e-2)

    def test__multiple_profiles__setup_fifth_weighted_sample_model__include_weight_and_likelihood(self, config_path,
                                                                                                  mn_samples_path):
        create_weighted_samples_10_parameters(mn_samples_path)

        nlo = non_linear.MultiNest(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=mn_samples_path)

        nlo.variable.light_profile = model_mapper.PriorModel(light_profiles.EllipticalExponential)
        nlo.variable.mass_profile = model_mapper.PriorModel(mass_profiles.SphericalNFW)

        nlo.save_model_info()

        weighted_sample_model, weight, likelihood = nlo.weighted_sample_instance_from_weighted_samples(index=5)

        assert weight == 0.1
        assert likelihood == -0.5 * 9999999.9

        assert weighted_sample_model.light_profile.centre == (1.0, 2.0)
        assert weighted_sample_model.light_profile.axis_ratio == 3.0
        assert weighted_sample_model.light_profile.phi == 4.0
        assert weighted_sample_model.light_profile.intensity == -5.0
        assert weighted_sample_model.light_profile.effective_radius == -6.0

        assert weighted_sample_model.mass_profile.centre == (-7.0, -8.0)
        assert weighted_sample_model.mass_profile.kappa_s == 9.0
        assert weighted_sample_model.mass_profile.scale_radius == 10.0

    def test__1_profile__limits_1d_vectors_via_weighted_samples__1d_vectors_are_correct(self, config_path,
                                                                                        mn_samples_path):
        create_weighted_samples_4_parameters(
            path=mn_samples_path)

        results = non_linear.MultiNest(
            model_mapper=model_mapper.ModelMapper(config=config.DefaultPriorConfig(config_path)),
            path=mn_samples_path)
        results.variable.mass_profile = mass_profiles.SphericalNFW

        results.save_model_info()

        assert results.model_at_upper_sigma_limit(sigma_limit=3.0) == pytest.approx([1.12, 2.12, 3.12, 4.12],
                                                                                    1e-2)
        assert results.model_at_lower_sigma_limit(sigma_limit=3.0) == pytest.approx([0.88, 1.88, 2.88, 3.88],
                                                                                    1e-2)


class MockAnalysis(object):
    def __init__(self):
        self.kwargs = None

    def fit(self, **kwargs):
        self.kwargs = kwargs
        return 1.


@pytest.fixture(name='test_config')
def make_test_config():
    return config.DefaultPriorConfig(
        config_folder_path="{}/../{}".format(os.path.dirname(os.path.realpath(__file__)),
                                             "test_files/config/priors/default"))


@pytest.fixture(name="width_config")
def make_width_config():
    return config.WidthConfig(
        config_folder_path="{}/../{}".format(os.path.dirname(os.path.realpath(__file__)),
                                             "test_files/config/priors/width"))


@pytest.fixture(name="downhill_simplex")
def make_downhill_simplex(test_config, width_config):
    def fmin(fitness_function, x0):
        fitness_function(x0)
        return x0

    return non_linear.DownhillSimplex(fmin=fmin, model_mapper=model_mapper.ModelMapper(config=test_config,
                                                                                       width_config=width_config))


@pytest.fixture(name="multi_nest")
def make_multi_nest(test_config, width_config, mn_samples_path):
    create_weighted_samples_4_parameters(mn_samples_path)

    # noinspection PyUnusedLocal
    def run(fitness_function, prior, total_parameters, outputfiles_basename, n_clustering_params=None,
            wrapped_params=None, importance_nested_sampling=True, multimodal=True, const_efficiency_mode=False,
            n_live_points=400, evidence_tolerance=0.5, sampling_efficiency=0.8, n_iter_before_update=100,
            null_log_evidence=-1e+90, max_modes=100, mode_tolerance=-1e+90, seed=-1, verbose=False, resume=True,
            context=0,
            write_output=True, log_zero=-1e+100, max_iter=0, init_MPI=False, dump_callback=None):
        fitness_function([1 for _ in range(total_parameters)], total_parameters, total_parameters, None)

    create_summary_4_parameters(mn_samples_path)
    return non_linear.MultiNest(run=run, path=mn_samples_path,
                                model_mapper=model_mapper.ModelMapper(config=test_config, width_config=width_config))


class TestFitting(object):
    class TestDownhillSimplex(object):
        def test_constant(self, downhill_simplex):
            downhill_simplex.constant.mock_class = MockClassNLOx4()
            result = downhill_simplex.fit(MockAnalysis())

            assert result.constant.mock_class.one == 1
            assert result.constant.mock_class.two == 2
            assert result.likelihood == 1

        def test_variable(self, downhill_simplex, test_config):
            downhill_simplex.variable.mock_class = model_mapper.PriorModel(MockClassNLOx4, test_config)
            result = downhill_simplex.fit(MockAnalysis())

            assert result.constant.mock_class.one == 0.5
            assert result.constant.mock_class.two == 0.5
            assert result.likelihood == 1

            assert result.variable.mock_class.one.mean == 0.5
            assert result.variable.mock_class.two.mean == 0.5

        def test_constant_and_variable(self, downhill_simplex, test_config):
            downhill_simplex.constant.constant = MockClassNLOx4()
            downhill_simplex.variable.variable = model_mapper.PriorModel(MockClassNLOx4, test_config)

            result = downhill_simplex.fit(MockAnalysis())

            assert result.constant.constant.one == 1
            assert result.constant.constant.two == 2
            assert result.constant.variable.one == 0.5
            assert result.constant.variable.two == 0.5
            assert result.variable.variable.one.mean == 0.5
            assert result.variable.variable.two.mean == 0.5
            assert result.likelihood == 1

    class TestMultiNest(object):
        def test_variable(self, multi_nest, test_config):
            multi_nest.variable.mock_class = model_mapper.PriorModel(MockClassNLOx4, test_config)
            result = multi_nest.fit(MockAnalysis())

            assert result.constant.mock_class.one == 9.0
            assert result.constant.mock_class.two == -10.0
            assert result.likelihood == 0.02

            assert result.variable.mock_class.one.mean == 1
            assert result.variable.mock_class.two.mean == -2


from autolens.profiles import geometry_profiles
from scipy.integrate import quad
import numpy as np
from itertools import count


class LightProfile(object):
    """Mixin class that implements functions common to all light profiles"""

    def intensity_at_radius(self, radius):
        """
        Abstract method for obtaining intensity at given radius
        Parameters
        ----------
        radius : float
            The distance from the centre of the profiles
        Returns
        -------
        intensity : float
            The value of intensity at the given radius
        """
        raise NotImplementedError("intensity_at_radius should be overridden")

    # noinspection PyMethodMayBeStatic
    def intensity_from_grid(self, grid):
        raise NotImplementedError("intensity_from_grid should be overridden")


# noinspection PyAbstractClass
class EllipticalLightProfile(geometry_profiles.EllipticalProfile, LightProfile):
    """Generic class for an elliptical light profiles"""

    _ids = count()

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalLightProfile, self).__init__(centre, axis_ratio, phi)
        self.axis_ratio = axis_ratio
        self.phi = phi

        self.component_number = next(self._ids)

    @property
    def subscript(self):
        return 'l'

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi']

    def luminosity_within_circle(self, radius):
        """
        Compute the light profiles's total luminosity within a circle of specified radius. This is performed via \
        integration and is centred on the light model_mapper.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the luminosity within.

        Returns
        -------
        luminosity : float
            The total luminosity within the specified circle.
        """
        return quad(self.luminosity_integral, a=0.0, b=radius, args=(1.0,))[0]

    def luminosity_within_ellipse(self, major_axis):
        """
        Compute the light profiles's total luminosity within an ellipse of specified major axis. This is performed via\
        integration and is centred, oriented and aligned with on the light model_mapper.
        Parameters
        ----------
        major_axis: float
            The major-axis of the ellipse to compute the luminosity within.
        Returns
        -------
        intensity : float
            The total luminosity within the specified ellipse.
        """
        return quad(self.luminosity_integral, a=0.0, b=major_axis, args=(self.axis_ratio,))[0]

    def luminosity_integral(self, x, axis_ratio):
        """Routine to integrate an elliptical light profiles - set axis ratio to 1 to compute the luminosity within a \
        circle"""
        r = x * axis_ratio
        return 2 * np.pi * r * self.intensity_at_radius(x)


class EllipticalSersic(EllipticalLightProfile):
    """The Sersic light profiles, used to fit and subtract the lens galaxy's light."""

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6,
                 sersic_index=4.0):
        """

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The circular radius containing half the light of this model_mapper
        sersic_index : Int
            The concentration of the light profiles
        """
        super(EllipticalSersic, self).__init__(centre, axis_ratio, phi)
        self.intensity = intensity
        self.effective_radius = effective_radius
        self.sersic_index = sersic_index

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', 'I', 'R', 'n']

    @property
    def elliptical_effective_radius(self):
        """The effective_radius term used in a Sersic light profiles is the circular effective radius. It describes the
         radius within which a circular aperture contains half the light profiles's light. For elliptical (i.e low axis
         ratio) systems, this circle won't robustly capture the light profiles's elliptical shape.

         The elliptical effective radius therefore instead describes the major-axis radius of the ellipse containing
         half the light, and may be more appropriate for pixelization of highly flattened systems like disk galaxies."""
        return self.effective_radius / np.sqrt(self.axis_ratio)

    @property
    def sersic_constant(self):
        """

        Returns
        -------
        sersic_constant: float
            A parameter, derived from sersic_index, that ensures that effective_radius always contains 50% of the light.
        """
        return (2 * self.sersic_index) - (1. / 3.) + (4. / (405. * self.sersic_index)) + (
                46. / (25515. * self.sersic_index ** 2)) + (131. / (1148175. * self.sersic_index ** 3)) - (
                       2194697. / (30690717750. * self.sersic_index ** 4))

    def intensity_at_radius(self, radius):
        """

        Parameters
        ----------
        radius : float
            The distance from the centre of the profiles
        Returns
        -------
        intensity : float
            The intensity at that distance
        """
        return self.intensity * np.exp(
            -self.sersic_constant * (((radius / self.effective_radius) ** (1. / self.sersic_index)) - 1))

    def intensity_at_grid_radii(self, grid_radii):
        return np.multiply(self.intensity, np.exp(
            np.multiply(-self.sersic_constant,
                        np.add(np.power(np.divide(grid_radii, self.effective_radius), 1. / self.sersic_index), -1))))

    @geometry_profiles.transform_grid
    def intensity_from_grid(self, grid):
        return self.intensity_at_grid_radii(self.grid_to_eccentric_radii(grid))


class EllipticalExponential(EllipticalSersic):
    """Used to fit flatter regions of light in a galaxy, typically a disk.

    It is a subset of the Sersic profiles, corresponding exactly to the reconstruction sersic_index = 1"""

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6):
        """

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The circular radius containing half the light of this model_mapper
        """
        super(EllipticalExponential, self).__init__(centre, axis_ratio, phi, intensity, effective_radius, 1.0)

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', 'I', 'R']


class EllipticalDevVaucouleurs(EllipticalSersic):
    """Used to fit the concentrated regions of light in a galaxy, typically its bulge. It may also fit the entire light
    profiles of an elliptical / early-type galaxy.

    It is a subset of the Sersic profiles, corresponding exactly to the reconstruction sersic_index = 4."""

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6):
        """

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The circular radius containing half the light of this model_mapper
        """
        super(EllipticalDevVaucouleurs, self).__init__(centre, axis_ratio, phi, intensity, effective_radius, 4.0)

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', 'I', 'R']


class EllipticalCoreSersic(EllipticalSersic):
    """The Core-Sersic profiles is used to fit the light of a galaxy. It is an extension of the Sersic profiles and \
    flattens the light profiles central values (compared to the extrapolation of a pure Sersic profiles), by forcing \
    these central regions to behave instead as a power-law."""

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6,
                 sersic_index=4.0, radius_break=0.01, intensity_break=0.05, gamma=0.25, alpha=3.0):
        """

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The circular radius containing half the light of this model_mapper
        sersic_index : Int
            The concentration of the light profiles
        radius_break : Float
            The break radius separating the inner power-law (with logarithmic slope gamma) and outer Sersic function.
        intensity_break : Float
            The intensity at the break radius.
        gamma : Float
            The logarithmic power-law slope of the inner core profiles
        alpha :
            Controls the sharpness of the transition between the inner core / outer Sersic profiles.
        """
        super(EllipticalCoreSersic, self).__init__(centre, axis_ratio, phi, intensity, effective_radius, sersic_index,
                                                   )
        self.radius_break = radius_break
        self.intensity_break = intensity_break
        self.alpha = alpha
        self.gamma = gamma

    @property
    def parameter_labels(self):
        return ['x', 'y', 'q', r'\phi', 'I', 'R', 'n', 'Rb', 'Ib', '\gamma', r'\alpha']

    @property
    def intensity_prime(self):
        """Overall intensity normalisation in the rescaled Core-Sersic light profiles (electrons per second)"""
        return self.intensity_break * (2.0 ** (-self.gamma / self.alpha)) * np.exp(
            self.sersic_constant * (((2.0 ** (1.0 / self.alpha)) * self.radius_break) / self.effective_radius) ** (
                    1.0 / self.sersic_index))

    def intensity_at_radius(self, radius):
        """

        Parameters
        ----------
        radius : float
            The distance from the centre of the profiles
        Returns
        -------
        intensity : float
            The intensity at that radius
        """
        return self.intensity_prime * (
                (1 + ((self.radius_break / radius) ** self.alpha)) ** (self.gamma / self.alpha)) * np.exp(
            -self.sersic_constant * ((((radius ** self.alpha) + (self.radius_break ** self.alpha)) / (
                    self.effective_radius ** self.alpha)) ** (
                                             1.0 / (self.alpha * self.sersic_index))))

    def intensity_at_grid_radii(self, grid_radii):
        return np.multiply(np.multiply(self.intensity_prime, np.power(
            np.add(1, np.power(np.divide(self.radius_break, grid_radii), self.alpha)), (self.gamma / self.alpha))),
                           np.exp(np.multiply(-self.sersic_constant,
                                              (np.power(np.divide(np.add(np.power(grid_radii, self.alpha), (
                                                      self.radius_break ** self.alpha)),
                                                                  (self.effective_radius ** self.alpha)), (
                                                                1.0 / (self.alpha * self.sersic_index)))))))



from autolens.pixelization import pixelization

import pytest
import numpy as np


# The regularization matrix, H, is calculated by defining a set of B matrices which describe how pix-plane
# pixels map to one another. For example, if we had a 3x3 square grid:

# ______
# |0|1|2|
# |3|4|5|
# |6|7|8|
# ^^^^^^^

# Lets say we want to regularize this grid so that each square pixel is regularized with a pixel to its right
# and below it.

# So, 0 is regularized with pixels 1 and 3, pixel 1 with 2 and 4, but pixel 2 with only pixel 5, etc. So,
#
# We make two 9 x 9 B matrices, which describe regularization in each direction. So for regularization to the
# right of each pixel:

# B_x = [-1,  1,  0,  0,  0,  0,  0,  0,  0] # [0->1] This, row 0, correspomds to pixel 0 (signified by the -1). The 1's in columns 1 is saying we want to regularize pixel 0 with pixel 1.
#       [ 0, -1,  1,  0,  0,  0,  0,  0,  0] # [1->2] Row 1 for pixel 1 (again, the -1 tells us this), regularized with pixels 2.
#       [ 0,  0, -1,  0,  0,  0,  0,  0,  0] # [] NOTE - pixel 2 is NOT regularized with pixel 3 (check the square grid)!
#       [ 0,  0,  0, -1,  1,  0,  0,  0,  0] # [3->4]
#       [ 0,  0,  0,  0, -1,  1,  0,  0,  0] # [4->5]
#       [ 0,  0,  0,  0,  0, -1,  0,  0,  0] # [] NOTE - pixel 5 not regularized with pixel 6!
#       [ 0,  0,  0,  0,  0,  0, -1,  1,  0] # [6->7]
#       [ 0,  0,  0,  0,  0,  0,  0, -1,  1] # [7->8]
#       [ 0,  0,  0,  0,  0,  0,  0,  0, -1] # [] NOTE - Not regularized with anything

# We now make another B matrix for the regularization beneath each pixel:

# B_y = [-1,  0,  0,  1,  0,  0,  0,  0,  0] # [0->3] This, row 0, correspodns to pixel 0 (signified by the -1). The 1's in columns 3 is saying we want to regularize pixel 0 with pixel 3.
#       [ 0, -1,  0,  0,  1,  0,  0,  0,  0] # [1->4] Row 1 for pixel 1 (again, the -1 tells us this), regularized with pixel 4
#       [ 0,  0, -1,  0,  0,  1,  0,  0,  0] # [2->5]
#       [ 0,  0,  0, -1,  0,  0,  1,  0,  0] # [3->6]
#       [ 0,  0,  0,  0, -1,  0,  0,  1,  0] # [4->7]
#       [ 0,  0,  0,  0,  0, -1,  0,  0,  1] # [5->8]
#       [ 0,  0,  0,  0,  0,  0, -1,  0,  0] # [] No regularized performed in these last 3 rows / pixels
#       [ 0,  0,  0,  0,  0,  0,  0, -1,  0] # []
#       [ 0,  0,  0,  0,  0,  0,  0,  0, -1] # []

# So, we basically just make B matrices representing regularization in each direction. For each, we
# can then compute their corresponding regularization matrix, H, as, H = B * B.T (matrix multiplication)

# So, H_x = B_x.T, * B_x H_y = B_y.T * B_y
# And our overall regularization matrix, H = H_x + H_y

# For an adaptive Voronoi grid, we do this, however we make a B matrix for every shared Voronoi vertex
# of each pix-pixel cluster. This means the number of B matrices we compute is equal to the the number of
# Voronoi vertices in the pix-pixel with the most Voronoi vertices (i.e. the most neighbours a pix-pixel has).

### COMBINING B MATRICES ###

# Whereas the paper_plots above had each -1 going down the diagonal, this is not necessary. It valid to put each pairing
# anywhere. So, if we had a 4x4 B matrix, where pixel 0 regularizes 1, 2 -> 3 and 3 -> 0, we can set this up
# as one matrix even though the pixel 0 comes up twice!

# B = [-1, 1, 0 ,0] # [0->1]
#     [0, 0, 0 ,0] # We can skip rows by making them all zeros.
#     [0, 0, -1 ,1] # [2->3]
#     [1, 0, 0 ,-1] # [3->0] This is valid!

# So, we don't have to make the same number of B matrices as Voronoi vertices, as we can combine them into a few B
# matrices like this

# SKIPPING THE B MATRIX CALCULATION #

# The create_regularization_matrix routines in pixelization don't use the B matrices to compute H!.

# This is because, if you know all the pairs between pix pixels (which the Voronoi gridding can tell you), you
# can bypass the B matrix multiplicaion entirely and enter the values directly into the H matrix (making the
# calculation significantly faster).

# POSITIVE DEFINITE MATRIX #

# The regularization matrix must be positive-definite, to ensure that its determinant is 0. To ensure this
# criteria is met, two forms of regularization schemes are applied in slightly different ways.

# CONSTANT REGULARIZATION #

# For the constant regularization scheme, there is only 1 regularization coefficient that is applied to all
# pix-pair regularizations equally. This means that we when write B, we only need to regularize pixels
# in one direction (e.g. pix pixel 0 regularizes pix pixel 1, but NOT visa versa). For example:

# B = [-1, 1]  [0->1]
#     [0, -1]  1 does not regularization with 0

# WEIGHTED REGULARIZATION #

# For the weighted regularization scheme, each pixis given an 'effective regularization weight', instead
# of applying one constant regularization coefficient to each pix-pixel pair. scheme overall. This is
# because different regions of a pix-plane want different levels of regularization
# (see Nightingale, Dye and Massey) 2018.

# Unlike the constant regularization scheme, regularization pairs are now defined such that all pixels are
# regularized with one another (e.g. if pix pixel 0 regularizes pix pixel 1, pix pixel 1 also
# regularizes pix pixel 0). For example :
#
# B = [-1, 1]  [0->1]
#     [-1, -1]  1 now also regularizes 0
#
# For a constant regularization coefficient this would NOT produce a positive-definite matrix. However, for
# the weighted scheme, it does!

# The regularize weights change the B matrix as shown below, we simply multiply each pix-pixel's effective
# regularization weight by each row of B it has a -1 in, so:

# regularization_weights = [1, 2, 3, 4]

# B = [-1, 1, 0 ,0] # [0->1]
#     [0, -2, 2 ,0] # [1->2]
#     [0, 0, -3 ,3] # [2->3]
#     [4, 0, 0 ,-4] # [3->0] This is valid!

# If our -1's werent down the diagonal this would look like:

# B = [4, 0, 0 ,-4] # [3->0]
#     [0, -2, 2 ,0] # [1->2]
#     [-1, 1, 0 ,0] # [0->1]
#     [0, 0, -3 ,3] # [2->3] This is valid!

class TestRegularizationConstant:

    class TestRegularizationMatrixFromNeighbors:

        def test__1_b_matrix_size_3x3__weights_all_1s__makes_correct_regularization_matrix(self):
            # Here, we define the pix_neighbors first here and make the B matrices based on them.

            # You'll notice that actually, the B Matrix doesn't have to have the -1's going down the diagonal and we
            # don't have to have as many B matrices as we do the pix pixel with the most  vertices. We can combine
            # the rows of each B matrix wherever we like ;0.

            pix_neighbors = np.array([[1, 2], [0], [0]])

            test_b_matrix = np.array([[-1, 1, 0],  # Pair 1
                                      [-1, 0, 1],  # Pair 2
                                      [0, 0, 0, ]])  # Pair 1 flip

            test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix) + 1e-8 * np.identity(3)

            pix = pixelization.ClusterRegConst(pixels=3, regularization_coefficients=(1.0,))
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()
            assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)

        def test__1_b_matrix_size_4x4__weights_all_1s__makes_correct_regularization_matrix(self):
            test_b_matrix = np.array([[-1, 1, 0, 0],
                                      [0, -1, 1, 0],
                                      [0, 0, -1, 1],
                                      [1, 0, 0, -1]])

            test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix) + 1e-8 * np.identity(4)

            pix_neighbors = np.array([[1, 3], [0, 2], [1, 3], [0, 2]])

            pix = pixelization.ClusterRegConst(pixels=4, regularization_coefficients=(1.0,))
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()
            assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)

        def test__1_b_matrix_size_4x4__coefficient_2__makes_correct_regularization_matrix(self):
            pix_neighbors = np.array([[1, 3], [0, 2], [1, 3], [0, 2]])

            test_b_matrix = 2.0 * np.array([[-1, 1, 0, 0],
                                            [0, -1, 1, 0],
                                            [0, 0, -1, 1],
                                            [1, 0, 0, -1]])

            test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix) + 1e-8 * np.identity(4)

            pix = pixelization.ClusterRegConst(pixels=4, regularization_coefficients=(2.0,))
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()
            assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)

        def test__1_b_matrix_size_9x9__coefficient_2__makes_correct_regularization_matrix(self):
            pix_neighbors = [[1, 3], [4, 2, 0], [1, 5], [4, 6, 0], [7, 1, 5, 3], [4, 2, 8], [7, 3], [4, 8, 6],
                             [7, 5]]

            test_b_matrix_0 = np.array([[-1, 1, 0, 0, 0, 0, 0, 0, 0],
                                        [-1, 0, 0, 1, 0, 0, 0, 0, 0],
                                        [0, -1, 1, 0, 0, 0, 0, 0, 0],
                                        [0, -1, 0, 0, 1, 0, 0, 0, 0],
                                        [0, 0, -1, 0, 0, 1, 0, 0, 0],
                                        [0, 0, 0, -1, 1, 0, 0, 0, 0],
                                        [0, 0, 0, -1, 0, 0, 1, 0, 0],
                                        [0, 0, 0, 0, -1, 1, 0, 0, 0],
                                        [0, 0, 0, 0, -1, 0, 0, 1, 0]])

            test_b_matrix_1 = np.array([[0, 0, 0, 0, 0, -1, 0, 0, 1],
                                        [0, 0, 0, 0, 0, 0, -1, 1, 0],
                                        [0, 0, 0, 0, 0, 0, 0, -1, 1],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0, 0, 0, 0]])

            test_regularization_matrix_0 = np.matmul(test_b_matrix_0.T, test_b_matrix_0)
            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

            test_regularization_matrix = test_regularization_matrix_0 + test_regularization_matrix_1 + 1e-8 * np.identity(
                9)

            pix = pixelization.ClusterRegConst(pixels=9, regularization_coefficients=(1.0,))
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()
            assert (abs(np.linalg.det(regularization_matrix)) > 1e-8)


class TestRegularizationWeighted:

    class TestComputeSourceSignals:

        def test__x3_image_pixels_signals_1s__pix_scale_1__pix_signals_all_1s(self):

            pix = pixelization.ClusterRegWeighted(pixels=3, pix_signal_scale=1.0)

            image_to_pix = np.array([0, 1, 2])
            galaxy_image = np.array([1.0, 1.0, 1.0])

            pix_signals = pix.pix_signals_from_images(image_to_pix, galaxy_image)

            assert (pix_signals == np.array([1.0, 1.0, 1.0])).all()

        def test__x4_image_pixels_signals_1s__pix_signals_still_all_1s(self):
            pix = pixelization.ClusterRegWeighted(pixels=3, pix_signal_scale=1.0)

            image_to_pix = np.array([0, 1, 2, 0])
            galaxy_image = np.array([1.0, 1.0, 1.0, 1.0])

            pix_signals = pix.pix_signals_from_images(image_to_pix, galaxy_image)

            assert (pix_signals == np.array([1.0, 1.0, 1.0])).all()

        def test__galaxy_flux_in_a_pix_pixel_is_double_the_others__pix_signal_is_1_others_a_half(self):
            pix = pixelization.ClusterRegWeighted(pixels=3, pix_signal_scale=1.0)

            image_to_pix = np.array([0, 1, 2])
            galaxy_image = np.array([2.0, 1.0, 1.0])

            pix_signals = pix.pix_signals_from_images(image_to_pix, galaxy_image)

            assert (pix_signals == np.array([1.0, 0.5, 0.5])).all()

        def test__same_as_above_but_pix_scale_2__scales_pix_signals(self):
            pix = pixelization.ClusterRegWeighted(pixels=3, pix_signal_scale=2.0)

            image_to_pix = np.array([0, 1, 2])
            galaxy_image = np.array([2.0, 1.0, 1.0])

            pix_signals = pix.pix_signals_from_images(image_to_pix, galaxy_image)

            assert (pix_signals == np.array([1.0, 0.25, 0.25])).all()

    class TestComputeRegularizationWeights(object):

        def test__pix_signals_all_1s__coefficients_all_1s__weights_all_1s(self):
            pix = pixelization.ClusterRegWeighted(pixels=3, regularization_coefficients=(1.0, 1.0))

            pix_signals = np.array([1.0, 1.0, 1.0])

            weights = pix.regularization_weights_from_pix_signals(pix_signals)

            assert (weights == np.array([1.0, 1.0, 1.0])).all()

        def test__pix_signals_vary__coefficents_all_1s__weights_still_all_1s(self):
            pix = pixelization.ClusterRegWeighted(pixels=3, regularization_coefficients=(1.0, 1.0))

            pix_signals = np.array([0.25, 0.5, 0.75])

            weights = pix.regularization_weights_from_pix_signals(pix_signals)

            assert (weights == np.array([1.0, 1.0, 1.0])).all()

        def test__pix_signals_vary__coefficents_1_and_0__weights_are_pix_signals_squared(self):
            pix = pixelization.ClusterRegWeighted(pixels=3, regularization_coefficients=(1.0, 0.0))

            pix_signals = np.array([0.25, 0.5, 0.75])

            weights = pix.regularization_weights_from_pix_signals(pix_signals)

            assert (weights == np.array([0.25 ** 2.0, 0.5 ** 2.0, 0.75 ** 2.0])).all()

        def test__pix_signals_vary__coefficents_0_and_1__weights_are_1_minus_pix_signals_squared(self):
            pix = pixelization.ClusterRegWeighted(pixels=3, regularization_coefficients=(0.0, 1.0))

            pix_signals = np.array([0.25, 0.5, 0.75])

            weights = pix.regularization_weights_from_pix_signals(pix_signals)

            assert (weights == np.array([0.75 ** 2.0, 0.5 ** 2.0, 0.25 ** 2.0])).all()

    class TestRegularizationMatrixFromNeighbors:

        def test__1_b_matrix_size_4x4__weights_all_1s__makes_correct_regularization_matrix(self):
            pix_neighbors = np.array([[2], [3], [0], [1]])

            test_b_matrix = np.array([[-1, 0, 1, 0],
                                      [0, -1, 0, 1],
                                      [1, 0, -1, 0],
                                      [0, 1, 0, -1]])

            test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix)

            regularization_weights = np.ones((4,))

            pix = pixelization.ClusterRegWeighted(pixels=4)
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(regularization_weights,
                                                                                          pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__2_b_matrices_size_3x3__weights_all_1s__makes_correct_regularization_matrix(self):
            # Here, we define the pix_neighbors first here and make the B matrices based on them.

            # You'll notice that actually, the B Matrix doesn't have to have the -1's going down the diagonal and we
            # don't have to have as many B matrices as we do the pix pixel with the most  vertices. We can combine
            # the rows of each B matrix wherever we like ;0.

            pix_neighbors = np.array([[1, 2], [0], [0]])

            test_b_matrix_1 = np.array([[-1, 1, 0],  # Pair 1
                                        [-1, 0, 1],  # Pair 2
                                        [1, -1, 0]])  # Pair 1 flip

            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

            test_b_matrix_2 = np.array([[1, 0, -1],  # Pair 2 flip
                                        [0, 0, 0],
                                        [0, 0, 0]])

            test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

            test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2

            regularization_weights = np.ones((3))

            pix = pixelization.ClusterRegWeighted(pixels=3)
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(regularization_weights,
                                                                                          pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__2_b_matrices_size_4x4__weights_all_1s__makes_correct_regularization_matrix(self):
            test_b_matrix_1 = np.array([[-1, 1, 0, 0],
                                        [0, -1, 1, 0],
                                        [0, 0, -1, 1],
                                        [1, 0, 0, -1]])

            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

            test_b_matrix_2 = np.array([[-1, 0, 0, 1],
                                        [1, -1, 0, 0],
                                        [0, 1, -1, 0],
                                        [0, 0, 1, -1]])

            test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

            test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2

            pix_neighbors = np.array([[1, 3], [0, 2], [1, 3], [0, 2]])
            regularization_weights = np.ones((4,))

            pix = pixelization.ClusterRegWeighted(pixels=4)
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(regularization_weights,
                                                                                          pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__4_b_matrices_size_6x6__weights_all_1s__makes_correct_regularization_matrix(self):
            # Again, lets exploit the freedom we have when setting up our B matrices to make matching it to pairs a
            # lot less Stressful.

            pix_neighbors = [[2, 3, 4], [2, 5], [0, 1, 3, 5], [0, 2], [5, 0], [4, 1, 2]]

            test_b_matrix_1 = np.array([[-1, 0, 1, 0, 0, 0],  # Pair 1
                                        [0, -1, 1, 0, 0, 0],  # Pair 2
                                        [-1, 0, 0, 1, 0, 0],  # Pair 3
                                        [0, 0, 0, 0, -1, 1],  # Pair 4
                                        [0, -1, 0, 0, 0, 1],  # Pair 5
                                        [-1, 0, 0, 0, 1, 0]])  # Pair 6

            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)

            test_b_matrix_2 = np.array([[0, 0, -1, 1, 0, 0],  # Pair 7
                                        [0, 0, -1, 0, 0, 1],  # Pair 8
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0]])

            test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

            test_b_matrix_3 = np.array([[1, 0, -1, 0, 0, 0],  # Pair 1 flip
                                        [0, 1, -1, 0, 0, 0],  # Pair 2 flip
                                        [1, 0, 0, -1, 0, 0],  # Pair 3 flip
                                        [0, 0, 0, 0, 1, -1],  # Pair 4 flip
                                        [0, 1, 0, 0, 0, -1],  # Pair 5 flip
                                        [1, 0, 0, 0, -1, 0]])  # Pair 6 flip

            test_regularization_matrix_3 = np.matmul(test_b_matrix_3.T, test_b_matrix_3)

            test_b_matrix_4 = np.array([[0, 0, 1, -1, 0, 0],  # Pair 7 flip
                                        [0, 0, 1, 0, 0, -1],  # Pair 8 flip
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0]])

            test_regularization_matrix_4 = np.matmul(test_b_matrix_4.T, test_b_matrix_4)

            test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2 + \
                                         test_regularization_matrix_3 + + test_regularization_matrix_4

            regularization_weights = np.ones((6))

            pix = pixelization.ClusterRegWeighted(pixels=6)
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(regularization_weights,
                                                                                          pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__1_b_matrix_size_3x3_variables_regularization_weights__makes_correct_regularization_matrix(self):
            # Simple case, where we have just one regularization direction, regularizing pixel 0 -> 1 and 1 -> 2.

            # This means our B matrix is:

            # [-1, 1, 0]
            # [0, -1, 1]
            # [0, 0, -1]

            # Regularization Matrix, H = B * B.T.I can

            regularization_weights = np.array([2.0, 4.0, 1.0])

            test_b_matrix = np.array([[-1, 1, 0],  # [[-2, 2, 0], (Matrix)
                                      [1, -1, 0],  # [4, -4, 0], (after)
                                      [0, 0, 0]])  # [0, 0,  0]]) (weights)

            test_b_matrix = (test_b_matrix.T * regularization_weights).T

            test_regularization_matrix = np.matmul(test_b_matrix.T, test_b_matrix)

            pix_neighbors = [[1], [0], []]

            pix = pixelization.ClusterRegWeighted(pixels=3)
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(regularization_weights,
                                                                                          pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__2_b_matrices_size_4x4_variables_regularization_weights__makes_correct_regularization_matrix(self):
            # Simple case, where we have just one regularization direction, regularizing pixel 0 -> 1 and 1 -> 2.

            # This means our B matrix is:

            # [-1, 1, 0]
            # [0, -1, 1]
            # [0, 0, -1]

            # Regularization Matrix, H = B * B.T.I can

            regularization_weights = np.array([2.0, 4.0, 1.0, 8.0])

            test_b_matrix_1 = np.array([[-2, 2, 0, 0],
                                        [-2, 0, 2, 0],
                                        [0, -4, 4, 0],
                                        [0, -4, 0, 4]])

            test_b_matrix_2 = np.array([[4, -4, 0, 0],
                                        [1, 0, -1, 0],
                                        [0, 1, -1, 0],
                                        [0, 8, 0, -8]])

            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)
            test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)

            test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2

            pix_neighbors = [[1, 2], [0, 2, 3], [0, 1], [1]]

            pix = pixelization.ClusterRegWeighted(pixels=4)
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(regularization_weights,
                                                                                          pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()

        def test__4_b_matrices_size_6x6_with_regularization_weights__makes_correct_regularization_matrix(self):
            pix_neighbors = [[1, 4], [2, 4, 0], [3, 4, 5, 1], [5, 2], [5, 0, 1, 2], [2, 3, 4]]
            regularization_weights = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])

            # I'm inputting the regularization weights directly thiss time, as it'd be a pain to multiply with a
            # loop.

            test_b_matrix_1 = np.array([[-1, 1, 0, 0, 0, 0],  # Pair 1
                                        [-1, 0, 0, 0, 1, 0],  # Pair 2
                                        [0, -2, 2, 0, 0, 0],  # Pair 3
                                        [0, -2, 0, 0, 2, 0],  # Pair 4
                                        [0, 0, -3, 3, 0, 0],  # Pair 5
                                        [0, 0, -3, 0, 3, 0]])  # Pair 6

            test_b_matrix_2 = np.array([[0, 0, -3, 0, 0, 3],  # Pair 7
                                        [0, 0, 0, -4, 0, 4],  # Pair 8
                                        [0, 0, 0, 0, -5, 5],  # Pair 9
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0]])

            # Now do the same pairs but with the regularization direction and weights swapped.

            test_b_matrix_3 = np.array([[2, -2, 0, 0, 0, 0],  # Pair 1
                                        [5, 0, 0, 0, -5, 0],  # Pair 2
                                        [0, 3, -3, 0, 0, 0],  # Pair 3
                                        [0, 5, 0, 0, -5, 0],  # Pair 4
                                        [0, 0, 4, -4, 0, 0],  # Pair 5
                                        [0, 0, 5, 0, -5, 0]])  # Pair 6

            test_b_matrix_4 = np.array([[0, 0, 6, 0, 0, -6],  # Pair 7
                                        [0, 0, 0, 6, 0, -6],  # Pair 8
                                        [0, 0, 0, 0, 6, -6],  # Pair 9
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0],
                                        [0, 0, 0, 0, 0, 0]])

            test_regularization_matrix_1 = np.matmul(test_b_matrix_1.T, test_b_matrix_1)
            test_regularization_matrix_2 = np.matmul(test_b_matrix_2.T, test_b_matrix_2)
            test_regularization_matrix_3 = np.matmul(test_b_matrix_3.T, test_b_matrix_3)
            test_regularization_matrix_4 = np.matmul(test_b_matrix_4.T, test_b_matrix_4)

            test_regularization_matrix = test_regularization_matrix_1 + test_regularization_matrix_2 + test_regularization_matrix_3 + test_regularization_matrix_4

            pix = pixelization.ClusterRegWeighted(pixels=6)
            regularization_matrix = pix.regularization_matrix_from_pix_neighbors(regularization_weights,
                                                                                          pix_neighbors)

            assert (regularization_matrix == test_regularization_matrix).all()



"""The hello command."""

from json import dumps

from .base import Base


class Hello(Base):
    """Say hello, world!"""

    def run(self):
        print('Hello, world!')
        print('You supplied the following options:', dumps(self.options, indent=2, sort_keys=True))



import numpy as np
from autolens.profiles import light_profiles as lps, mass_profiles as mps
from itertools import count


def is_light_profile(obj):
    return isinstance(obj, lps.LightProfile) and not isinstance(obj, mps.MassProfile)


def is_mass_profile(obj):
    return isinstance(obj, mps.MassProfile)


class Galaxy(object):
    """
    @DynamicAttrs
    """

    def __init__(self, redshift=None, pixelization=None, hyper_galaxy=None, **kwargs):
        """
        Represents a real galaxy. This could be a lens galaxy or source galaxy. Note that a lens galaxy must have mass
        profiles

        Parameters
        ----------
        redshift: float
            The redshift of this galaxy
        light_profiles: [LightProfile]
            A list of light profiles describing the light profiles of this galaxy
        mass_profiles: [MassProfile]
            A list of mass profiles describing the mass profiles of this galaxy
        """
        self.redshift = redshift

        for name, val in kwargs.items():
            setattr(self, name, val)

        self.pixelization = pixelization
        self.hyper_galaxy = hyper_galaxy

    @property
    def light_profiles(self):
        return [value for value in self.__dict__.values() if is_light_profile(value)]

    @property
    def mass_profiles(self):
        return [value for value in self.__dict__.values() if is_mass_profile(value)]

    @property
    def has_pixelization(self):
        """
        Returns
        -------
        True iff this galaxy has an associated pixelization
        """
        return self.pixelization is not None

    @property
    def has_hyper_galaxy(self):
        """
        Returns
        -------
        True iff this galaxy has an associated hyper galaxy
        """
        return self.hyper_galaxy is not None

    @property
    def has_profile(self):
        """
        Returns
        -------
        True iff there is one or more mass or light profiles associated with this galaxy
        """
        return len(self.mass_profiles) + len(self.light_profiles) > 0

    def __repr__(self):
        string = "Redshift: {}".format(self.redshift)
        if self.pixelization:
            string += str(self.pixelization)
        if self.hyper_galaxy:
            string += str(self.hyper_galaxy)
        if self.light_profiles:
            string += "\nLight Profiles:\n{}".format("\n".join(map(str, self.light_profiles)))
        if self.mass_profiles:
            string += "\nMass Profiles:\n{}".format("\n".join(map(str, self.mass_profiles)))
        return string

    def intensity_from_grid(self, grid):
        if self.light_profiles is not None and len(self.light_profiles) > 0:
            return sum(map(lambda p: p.intensity_from_grid(grid), self.light_profiles))
        else:
            return np.zeros((grid.shape[0],))

    def intensity_from_grid_individual(self, grid):
        """
        Compute the individual intensities of the galaxy's light profiles at a given set of image_grid.

        See *light_profiles* module for details of how this is performed.

        Parameters
        ----------
        grid : ndarray
            The image_grid in image_grid space
        Returns
        -------
        intensity : [float]
            The summed values of intensity at the given image_grid
        """
        return list(map(lambda p: p.intensity_from_grid(grid), self.light_profiles))

    def luminosity_within_circle(self, radius):
        """
        Compute the total luminosity of the galaxy's light profiles within a circle of specified radius.

        See *light_profiles.luminosity_within_circle* for details of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the luminosity within.

        Returns
        -------
        luminosity : float
            The total combined luminosity within the specified circle.
        """
        return sum(map(lambda p: p.luminosity_within_circle(radius), self.light_profiles))

    def luminosity_within_circle_individual(self, radius):
        """
        Compute the individual total luminosity of each light profile in the galaxy, within a circle of
        specified radius.

        See *light_profiles.luminosity_within_circle* for details of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the luminosity within.

        Returns
        -------
        luminosity : [float]
            The total combined luminosity within the specified circle.
        """
        return list(map(lambda p: p.luminosity_within_circle(radius), self.light_profiles))

    def luminosity_within_ellipse(self, major_axis):
        """
        Compute the total luminosity of the galaxy's light profiles, within an ellipse of specified major axis. This 
        is performed via integration of each light profile and is centred, oriented and  aligned with each light 
        model_mapper's individual geometry.

        See *light_profiles.luminosity_within_ellipse* for details of how this is performed.

        Parameters
        ----------
        major_axis: float
            The major-axis of the ellipse to compute the luminosity within.
        Returns
        -------
        intensity : float
            The total luminosity within the specified ellipse.
        """
        return sum(map(lambda p: p.luminosity_within_ellipse(major_axis), self.light_profiles))

    def luminosity_within_ellipse_individual(self, major_axis):
        """
        Compute the individual total luminosity of each light profile in the galaxy, within an ellipse of 
        specified major axis.

        See *light_profiles.luminosity_within_ellipse* for details of how this is performed.

        Parameters
        ----------
        major_axis: float
            The major-axis of the ellipse to compute the luminosity within.
        Returns
        -------
        intensity : [float]
            The total luminosity within the specified ellipse.
        """
        return list(map(lambda p: p.luminosity_within_ellipse(major_axis), self.light_profiles))

    def surface_density_from_grid(self, grid):
        """

        Compute the summed surface density of the galaxy's mass profiles at a given set of image_grid.

        See *mass_profiles* module for details of how this is performed.

        Parameters
        ----------
        grid : ndarray
            The x and y image_grid of the image_grid

        Returns
        ----------
        The summed values of surface density at the given image_grid.
        """
        return sum(map(lambda p: p.surface_density_from_grid(grid), self.mass_profiles))

    def surface_density_from_grid_individual(self, grid):
        """

        Compute the individual surface densities of the galaxy's mass profiles at a given set of image_grid.

        See *mass_profiles* module for details of how this is performed.

        Parameters
        ----------
        grid : (float, float)
            The x and y image_grid of the image_grid

        Returns
        ----------
        The summed values of surface density at the given image_grid.
        """
        return list(map(lambda p: p.surface_density_from_grid(grid), self.mass_profiles))

    def potential_from_grid(self, grid):
        """
        Compute the summed gravitational potential of the galaxy's mass profiles at a given set of image_grid.

        See *mass_profiles* module for details of how this is performed.

        Parameters
        ----------
        grid : ndarray
            The x and y image_grid of the image_grid

        Returns
        ----------
        The summed values of gravitational potential at the given image_grid.
        """
        return sum(map(lambda p: p.potential_from_grid(grid), self.mass_profiles))

    def potential_from_grid_individual(self, grid):
        """
        Compute the individual gravitational potentials of the galaxy's mass profiles at a given set of image_grid.

        See *mass_profiles* module for details of how this is performed.

        Parameters
        ----------
        grid : Union(ndarray, (float, float))
            The x and y image_grid of the image_grid

        Returns
        ----------
        The summed values of gravitational potential at the given image_grid.
        """
        return list(map(lambda p: p.potential_from_grid(grid), self.mass_profiles))

    def deflections_from_grid(self, grid):
        if self.mass_profiles is not None and len(self.mass_profiles) > 0:

            deflections = sum(map(lambda p: p.deflections_from_grid(grid), self.mass_profiles))

            # return sum(map(lambda p: p.deflections_from_coordinate_grid(grid), self.mass_profiles))
            return deflections
        else:
            return np.full((grid.shape[0], 2), 0.0)

    def deflections_from_grid_individual(self, grid):
        """
        Compute the individual deflection angles of the galaxy's mass profiles at a given set of image_grid.

        See *mass_profiles* module for details of how this is performed.

        Parameters
        ----------
        grid : Union(np.ndarray, (float, float))
            The x and y image_grid of the image_grid

        Returns
        ----------
        The summed values of deflection angles at the given image_grid.
        """
        return np.asarray(list(map(lambda p: p.deflections_from_grid(grid), self.mass_profiles)))

    def dimensionless_mass_within_circles(self, radius):
        """
        Compute the total dimensionless mass of the galaxy's mass profiles within a circle of specified radius.

        See *mass_profiles.dimensionless_mass_within_circles* for details of how this is performed.


        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.

        Returns
        -------
        dimensionless_mass : float
            The total dimensionless mass within the specified circle.
        """
        return sum(map(lambda p: p.dimensionless_mass_within_circle(radius), self.mass_profiles))

    def dimensionless_mass_within_circles_individual(self, radius):
        """
        Compute the individual dimensionless mass of the galaxy's mass profiles within a circle of specified radius.

        See *mass_profiles.dimensionless_mass_within_circles* for details of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.

        Returns
        -------
        dimensionless_mass : ndarray
            The total dimensionless mass within the specified circle.
        """
        return np.asarray(list(map(lambda p: p.dimensionless_mass_within_circle(radius), self.mass_profiles)))

    def dimensionless_mass_within_ellipses(self, major_axis):
        """
        Compute the total dimensionless mass of the galaxy's mass profiles within an ellipse of specified major_axis.

        See *mass_profiles.dimensionless_mass_within_ellipses* for details of how this is performed.


        Parameters
        ----------
        major_axis : float
            The major axis of the ellipse

        Returns
        -------
        dimensionless_mass : float
            The total dimensionless mass within the specified circle.
        """
        return sum(map(lambda p: p.dimensionless_mass_within_ellipse(major_axis), self.mass_profiles))

    def dimensionless_mass_within_ellipses_individual(self, major_axis):
        """
        Compute the individual dimensionless mass of the galaxy's mass profiles within an ellipse of specified 
        major-axis.

        See *mass_profiles.dimensionless_mass_within_circles* for details of how this is performed.

        Parameters
        ----------
        major_axis : float
            The major axis of the ellipse

        Returns
        -------
        dimensionless_mass : ndarray
            The total dimensionless mass within the specified circle.
        """
        return list(map(lambda p: p.dimensionless_mass_within_ellipse(major_axis), self.mass_profiles))


# TODO : Should galaxy image and minimum value be in the constructor (they aren't free parameters)?

class HyperGalaxy(object):
    _ids = count()

    def __init__(self, contribution_factor=0.0, noise_factor=0.0, noise_power=1.0):
        """Class for scaling the noise in the different galaxies of an image (e.g. the lens, source).

        Parameters
        -----------
        contribution_factor : float
            Factor that adjusts how much of the galaxy's light is attributed to the contribution map.
        noise_factor : float
            Factor by which the noise is increased in the regions of the galaxy's contribution map.
        noise_power : float
            The power to which the contribution map is raised when scaling the noise.
        """
        self.contribution_factor = contribution_factor
        self.noise_factor = noise_factor
        self.noise_power = noise_power

        self.component_number = next(self._ids)

    @property
    def subscript(self):
        return 'hg'

    @property
    def parameter_labels(self):
        return [r'\Omega', r'\omega1', r'\omega2']

    def contributions_from_model_images(self, model_image, galaxy_image, minimum_value):
        """Compute the contribution map of a galaxy, which represents the fraction of flux in each pixel that \
        galaxy can be attributed to contain.

        This is computed by dividing that galaxy's flux by the total flux in that pixel, and then scaling by the \
        maximum flux such that the contribution map ranges between 0 and 1.

        Parameters
        -----------
        model_image : ndarray
            The model image of the observed weighted_data (from a previous analysis phase). This tells us the total light \
            attributed to each image pixel by the model.
        galaxy_image : ndarray
            A model image of the galaxy (e.g the lens light profile or source reconstruction) computed from a
            previous analysis.
        minimum_value : float
            The minimum fractional flux a pixel must contain to not be rounded to 0.
        """
        contributions = galaxy_image / (model_image + self.contribution_factor)
        contributions = contributions / np.max(contributions)
        contributions[contributions < minimum_value] = 0.0
        return contributions

    def scaled_noise_for_contributions(self, noise, contributions):
        """Compute a scaled galaxy noise map from a baseline nosie map.

        This uses the galaxy contribution map with their noise scaling hyper-parameters.

        Parameters
        -----------
        noise : ndarray
            The noise before scaling (this may already have the background scaled in HyperImage)
        contributions : ndarray
            The galaxy contribution map.
        """
        return self.noise_factor * (noise * contributions) ** self.noise_power

    def __eq__(self, other):
        if isinstance(other, HyperGalaxy):
            return self.contribution_factor == other.contribution_factor and \
                   self.noise_factor == other.noise_factor and \
                   self.noise_power == other.noise_power
        return False

    def __str__(self):
        return "\n".join(["{}: {}".format(k, v) for k, v in self.__dict__.items()])


class Redshift(object):
    def __init__(self, redshift):
        self.redshift = redshift

    def __str__(self):
        return str(self.redshift)

    @property
    def parameter_labels(self):
        return 'z'

    @property
    def subscript(self):
        return 'g'
